"""
Interface for variable management.
"""

from copy import deepcopy
import functools
import logging
from typing import List, Iterable

import idc
import ida_bytes
import ida_struct
import ida_typeinf

from kordesii.utils.function_tracing import utils


logger = logging.getLogger(__name__)


class VariableMap(object):
    """
    Class that stores a set of variables that have been encountered during emulation.
    """

    def __init__(self, cpu_context):
        self._variables = {}
        self._cpu_context = cpu_context

    def __repr__(self):
        return "<VariableMap : \n\t{}\n>".format(
            "\n\t".join(([repr(var) for addr, var in sorted(self._variables.items())]))
        )

    def __deepcopy__(self, memo):
        """
        Custom implementation of deepcopy to improve efficiency.
        """
        copy = VariableMap(deepcopy(self._cpu_context, memo))
        memo[id(self)] = copy
        copy._variables = {addr: deepcopy(variable, memo) for addr, variable in list(self._variables.items())}
        return copy

    def __getitem__(self, addr_or_name) -> "Variable":
        """Gets a variable by name or address."""
        if isinstance(addr_or_name, str):
            name = addr_or_name
            for var in self:
                if name == var.name:
                    return var
            raise KeyError(f"{name} not found.")
        elif isinstance(addr_or_name, int):
            return self._variables[addr_or_name]
        else:
            raise ValueError("Invalid variable name or address: {!r}".format(addr_or_name))

    def __len__(self):
        return len(self._variables)

    def get(self, addr_or_name, default=None) -> "Variable":
        """Gets a variable by name or address."""
        try:
            return self[addr_or_name]
        except (KeyError, ValueError):
            return default

    def at(self, ip) -> List["Variable"]:
        """
        Retrieves the variables referenced at the given instruction address.

        :param ip: Instruction address to get pointers from.
        :return: List of Variable objects that were found within the given instruction.

        :raises ValueError: If instruction has not been executed yet.
        """
        if ip not in self._cpu_context.executed_instructions:
            raise ValueError("Unable to get variables. Instruction at 0x{:0x} has not been executed.".format(ip))

        return [var for var in self if ip in var.references]

    def __setitem__(self, addr_or_name, variable):
        """Sets a variable by name or address."""
        if isinstance(addr_or_name, str):  # TODO
            raise NotImplementedError("Creating new variable by name is currently not supported.")
        elif isinstance(addr_or_name, int):
            self._variables[addr_or_name] = variable
        else:
            raise ValueError("Invalid variable name or address: {!r}".format(addr_or_name))

    def __contains__(self, addr_or_name) -> bool:
        if isinstance(addr_or_name, str):
            name = addr_or_name
            return any(name == var.name for var in self)
        elif isinstance(addr_or_name, int):
            addr = addr_or_name
            return addr in self._variables
        else:
            raise ValueError("Invalid variable name or address: {!r}".format(addr_or_name))

    def __iter__(self) -> Iterable["Variable"]:
        return iter(self._variables.values())

    def add(self, addr, frame_id=None, stack_offset=None, reference=None) -> "Variable":
        """
        Creates and adds a variable object to mapping by object

        If the variable already exists, this function does nothing.

        :return: Variable object that has been created or one that already exists.
        """
        if addr in self._variables:
            var = self._variables[addr]
        else:
            var = Variable(self._cpu_context, addr, frame_id=frame_id, stack_offset=stack_offset)
            # logger.debug('VariableMap :: Created variable: {!r}'.format(var))
            self._variables[addr] = var
        if reference:
            var.add_reference(reference)
        return var

    @property
    def names(self) -> List[str]:
        return [var.name for var in self]

    @property
    def addrs(self) -> List[int]:
        return list(self._variables.keys())

    @property
    def stack_variables(self) -> List["Variable"]:
        return [var for var in self if var.is_stack]

    @property
    def global_variables(self) -> List["Variable"]:
        return [var for var in self if not var.is_stack]


@functools.total_ordering
class Variable(object):
    """
    Stores information for a local / global variable for a specific CPU context state.
    """

    TYPE_MAP = {
        idc.FF_BYTE: "byte",
        idc.FF_WORD: "word",
        idc.FF_DWORD: "dword",
        idc.FF_QWORD: "qword",
        idc.FF_OWORD: "oword",
        idc.FF_TBYTE: "tbyte",
        idc.FF_STRLIT: "char",
        idc.FF_STRUCT: "struct",
        idc.FF_FLOAT: "float",
        idc.FF_DOUBLE: "double",
        idc.FF_PACKREAL: "packed decimal real",
        idc.FF_ALIGN: "alignment directive",
    }

    SIZE_MAP = {
        idc.FF_WORD: 2,
        idc.FF_DWORD: 4,
        idc.FF_QWORD: 8,
        idc.FF_OWORD: 16,
    }

    def __init__(self, cpu_context, addr, frame_id=None, stack_offset=None):
        if (frame_id is not None and stack_offset is None) or (frame_id is None and stack_offset is not None):
            raise ValueError("Both frame_id and stack_offset must be provided.")
        self._cpu_context = cpu_context
        self.addr = addr
        self.frame_id = frame_id
        self.stack_offset = stack_offset
        self.references = []  # list of instruction pointers where the variable was encountered.

    def __deepcopy__(self, memo):
        copy = self.__new__(self.__class__)
        memo[id(self)] = copy
        copy._cpu_context = deepcopy(self._cpu_context, memo)
        copy.addr = self.addr
        copy.frame_id = self.frame_id
        copy.stack_offset = self.stack_offset
        copy.references = list(self.references)
        return copy

    def __repr__(self):
        data_type_str = self.data_type
        if self.count > 1 and data_type_str != "func_ptr":
            data_type_str += f"[{self.count}]"
        string = (
            f"<Variable {self.name} "
            f": type = {data_type_str} "
            f": addr = 0x{self.addr:0x} "
            f": value = {repr(self.value)} "
            f": size = {self.size} "
        )
        if self.is_stack:
            string += f": frame_id = 0x{self.frame_id:0x} : stack_offset = {self.stack_offset} "
        string += ">"
        return string

    def __eq__(self, other):
        return self.addr == other.addr

    def __lt__(self, other):
        return self.addr < other.addr

    @property
    def _struc(self):
        return ida_struct.get_struc(self.frame_id)

    @property
    def _member(self):
        return ida_struct.get_member(self._struc, self.stack_offset)

    @property
    def is_stack(self):
        """True if variable is on stack."""
        return self.stack_offset is not None

    @property
    def is_func_ptr(self):
        return utils.is_func_ptr(self.addr)

    @property
    def name(self):
        if self.is_stack:
            name = ida_struct.get_member_name(self._member.id)
        else:
            name = idc.get_name(self.addr)
        if not name:
            return ""
        return name

    @property
    def size(self):
        """Size of data"""
        if self.is_stack:
            return ida_struct.get_member_size(self._member)
        else:
            return ida_bytes.get_item_size(self.addr)

    @property
    def data_type_size(self) -> int:
        """The data type size, defaults to 1 if unknown"""
        if self.is_stack:
            tif = ida_typeinf.tinfo_t()
            success = ida_struct.get_member_tinfo(tif, self._member)
            if not success:
                # Sometimes IDA will fail to get member type information for unknown reasons.
                # In these cases, set the type size ourselves if it is obvious or default to 1.
                return self.SIZE_MAP.get(self._data_type_enum, 1)
            return tif.get_size()
        else:
            return ida_bytes.get_data_elsize(self.addr, self._data_type_enum)

    @property
    def count(self):
        """Count of elements in the array."""
        return self.size // self.data_type_size

    @property
    def data(self):
        """The raw data the variable is pointing to."""
        return self._cpu_context.mem_read(self.addr, self.size)

    @data.setter
    def data(self, value):
        """Sets the raw data the variable is pointing to."""
        size = self.size
        if len(value) > size:
            raise ValueError(f"Data size for variable at 0x{self.addr:08x} ({self.name}) must be <= {size} bytes.")

        self._cpu_context.mem_write(self.addr, value)

    @property
    def _data_type_enum(self) -> int:
        """The data type as a IDA enum"""
        if self.is_stack:
            flags = self._member.flag
        else:
            flags = ida_bytes.get_flags(self.addr)
        return flags & idc.DT_TYPE

    @property
    def data_type(self) -> str:
        """The data type as a string."""
        if self.is_func_ptr:
            return "func_ptr"
        else:
            return self.TYPE_MAP.get(self._data_type_enum, "")

    def add_reference(self, ip):
        """Adds ip to list of references for this variable."""
        # Ignore duplicate calls.
        if self.references and ip == self.references[-1]:
            return
        self.references.append(ip)

    def _data_array(self) -> List[int]:
        """Returns data as an array of unpacked integers based on data_type size."""
        data = self.data
        data_type_size = self.data_type_size
        return [utils.struct_unpack(data[i:i + data_type_size]) for i in range(0, len(data), data_type_size)]

    @property
    def value(self):
        """The unpacked data the variable is pointing to."""
        data_type = self.data_type

        if data_type == "func_ptr":
            return self.addr

        if data_type in ("char", "byte", "tbyte"):
            return self.data

        if data_type in ("word", "dword", "qword", "oword"):
            data_array = self._data_array()
            if len(data_array) == 1:
                return data_array[0]
            return data_array

        if data_type in ("float", "double"):
            data_array = [utils.int_to_float(value) for value in self._data_array()]
            if len(data_array) == 1:
                return data_array[0]
            return data_array

        if data_type == "struct":
            # TODO: Support returning some type of dictionary for structs.
            return self.data

        raise NotImplementedError(f"Unsupported data type: {data_type}")

    @value.setter
    def value(self, value):
        """Set the data the variable is pointing to."""
        data_type = self.data_type

        if data_type in ("char", "tbyte"):
            self.data = value

        elif data_type in ("byte", "word", "dword", "qword", "oword"):
            width = self.data_type_size
            if isinstance(value, list):
                self.data = b''.join(utils.struct_pack(_value, width=width) for _value in value)
            else:
                self.data = utils.struct_pack(value, width=width)

        elif data_type in ("float", "double"):
            width = self.data_type_size
            if isinstance(value, list):
                self.data = b''.join(
                    utils.struct_pack(utils.float_to_int(_value), width=width) for _value in value)
            else:
                self.data = utils.struct_pack(utils.float_to_int(value), width=width)

        else:
            raise NotImplementedError(f"Unsupported data type: {data_type}")

    @property
    def history(self):
        """The history of variables by following memory copies."""
        # (We shouldn't have Nones)
        history = []
        for addr in self._cpu_context.get_pointer_history(self.addr):
            var = self._cpu_context.variables.get(addr, None)
            if var:
                history.append(var)
        return history



import abc
import copy
import hashlib
import logging
import os
import warnings

import idaapi
import idautils
import idc

logger = logging.getLogger(__name__)

# region ======== Deprecated Components ========

import kordesii.utils.strings as strings
import kordesii.utils.functions as functions
import kordesii.utils.utils as utils
import kordesii.utils.ida_re as ida_re
import kordesii.utils.yara as yara


class EncodedString(strings.EncodedString):
    def __init__(self, *args, **kwargs):
        warnings.warn("EncodedString has been moved to kordesii.utils.EncodedString", DeprecationWarning)
        super(EncodedString, self).__init__(*args, **kwargs)


class EncodedStackString(strings.EncodedStackString):
    def __init__(self, *args, **kwargs):
        warnings.warn("EncodedStackString has been moved to kordesii.utils.EncodedStackString", DeprecationWarning)
        super(EncodedStackString, self).__init__(*args, **kwargs)


class SuperFunc_t(functions.Function):
    def __init__(self, *args, **kwargs):
        warnings.warn("SuperFunc_t has been moved and renamed to kordesii.utils.Function", DeprecationWarning)
        super(SuperFunc_t, self).__init__(*args, **kwargs)


def find_destination(*args, **kwargs):
    warnings.warn("find_destination has moved to kordesii.utils.find_destination", DeprecationWarning)
    utils.find_destination(*args, **kwargs)


def re_find_functions(regex, flags=0, section=None, func_name=None):
    warnings.warn(
        "re_find_functions has been moved to kordesii.utils.ida_re.find_functions",
        DeprecationWarning
    )
    funcs = list(ida_re.find_functions(regex, flags=flags, segname=section))
    if func_name is not None:
        for func in funcs:
            func.rename(func_name)
    return funcs


def yara_find_decode_functions(rule_text, func_name=None):
    warnings.warn(
        "yara_find_decode_functions has been moved to kordesii.utils.yara.find_functions",
        DeprecationWarning
    )
    funcs = list(yara.find_functions(rule_text))
    if func_name is not None:
        for func in funcs:
            func.rename(func_name)
    return funcs


def make_superfunc_t_from_matches(matches, func_name=None):
    warnings.warn("make_superfunc_t_from_matches() is deprecated.", DeprecationWarning)
    decode_funcs = set()
    for ea, identifier in matches:
        if ea == idc.BADADDR:
            continue
        func = functions.Function(ea, identifier)
        if func_name is not None:
            func.rename(func_name)
        decode_funcs.add(func)
    return list(decode_funcs)

# endregion =====================================


class StringTracer(object, metaclass=abc.ABCMeta):
    """
    Description:
        An object to hold tracing info for one offset at a time.

        This abstract class doesn't do anything, but attempts to define fields that are
        almost always used. If a field is not used, it is recommended to set it to UNUSED.

    Fields:
        initial_offset - The EA the searching starts at. This is usually a yara match EA or a
                         func_t.start_ea. (<initial_offset> must be within a function.)
        func_ea - The start_ea of the function containing the initial_offset.
        string_location - The EA the encoded string starts at. Defaults to idc.BADADDR.
        string_reference - The EA from which the encoded string is referenced. Defaults to
                           idc.BADADDR.
        encoded_strings - Storage for the encoded string(s) found by search. Defaults to [].
        size - The size of the string. Defaults to None.

    Input:
        initial_offset - Required. We have to start somewhere.
        identifier - The id of the YARA rule that this Tracer is based off of

    Throws:
        AttributeError - There was no function at initial_offset.
    """

    def __init__(self, initial_offset=None, identifier=None):
        super(StringTracer, self).__init__()
        self.initial_offset = initial_offset
        self.identifier = identifier
        if self.initial_offset is not None:
            if not idaapi.get_func(initial_offset):
                raise AttributeError("No function at 0x%X" % initial_offset)
            self.func_ea = idaapi.get_func(initial_offset).start_ea
        else:
            self.func_ea = None
        # TODO: remove use of idc.BADADDR
        self.string_location = idc.BADADDR
        self.string_reference = idc.BADADDR
        self.encoded_strings = []
        self.size = None

    @abc.abstractmethod
    def search(self):
        """
        Description:
            Attempts to identify the string location based on the address passed in the constructor.

            A reminder: abc doesn't care about your parameters matching this prototype, just the name,
            so using **kwargs on search is fine. However, because find_encoded_strings is calling
            this function, there may be no *args as it wouldn't know what to pass. Assuming you don't
            edit this file's code, it may be useful to simply wrap your actual search function with this
            search function in some cases.

        Output:
            True if the encoded string offset is identified, False if the encoded string offset
            is not identified.
        """
        pass


# Put these here for increased robustness. Please don't depend on these very often.
ENCODED_STRINGS = []
DECODED_STRINGS = []


def get_encoded_strings():
    """ Use responsibly """
    return ENCODED_STRINGS


def get_decoded_strings():
    """ Use responsibly """
    return DECODED_STRINGS


def is_valid_ea(ea):
    """
    Description:
        Returns true for valid EAs, False for invalid ones.

    Input:
        ea - The EA to check

    Output:
        True if the EA is valid, False if it is not
    """
    return ea not in (None, idc.BADADDR) and idc.get_inf_attr(
        idc.INF_MIN_EA
    ) <= ea <= idc.get_inf_attr(idc.INF_MAX_EA)


def split_decoded_string(decoded_string, identify=False):
    """
    Description:
        Given a single EncodedString, split it into multiple.
        By default, it is split on \x00.
        The original EncodedString is not modified.

    Input:
        decoded_string - The EncodedString to split
        identify - When True, define the resultant strings in the IDB

    Output:
        A list of EncodedStrings.
    """
    """
    Normally, we have either a block of either all one byte characters OR all two byte characters,
    however with some families, the blocks have utf-8 and utf-16 mixed together. When the block is all one
    byte characters, we just split the block on \x00 and are done with it. Likewise, when the block is all
    two byte characters, we just split the block on \x00\x00 and are done with it. For mixed blocks,
    we can't just split on one or the other. If we were to split on \x00 to find the single byte character strings,
    we would of course completely destroy the two byte character strings. Therefore, we effectively tokenize
    the block on \x00\x00 and process each sub-block separately.

    To process a block, we start from its end (which is guaranteed to be either \x00\x00 or the end of
    the buffer, which might as well be \x00) and work our way forward. We initially start by finding
    the rightmost \x00, which is either the end of the previous single byte character string or the
    first byte in a two byte character. To determine if the \x00 is part of two byte character string,
    we step back two characters. We keep stepping back as long as we keep landing
    on \x00. Once we hit a character that's not \x00, we undo our last step (and because we don't want
    to grab a leading \x00, step one more character, for a total of 3 characters not the 2 we have
    been stepping).

    Wherever we are now is the beginning of the string, which continues up until the index from which
    we searched for the rightmost \x00 earlier. We continue like this until we run into the front of
    the current block and then process the next block, repeating until we hit the end of the buffer.
    """
    warnings.warn(
        "split_decoded_string() is deprecated. "
        "Please use .split() on the EncodedString object instead.",
        DeprecationWarning
    )

    results = []
    section_start = 0
    section_end = decoded_string.decoded_data.find(b"\x00\x00")
    if section_end == -1:
        section_end = len(decoded_string.encoded_data)
    string_end = section_end

    while True:
        while True:
            # Determine where the next individual string starts
            string_start = decoded_string.decoded_data[section_start:string_end].rfind(
                b"\x00"
            )
            if string_start != -1:
                string_start += section_start
                while (
                    string_start >= section_start
                    and decoded_string.decoded_data[string_start] == 0
                ):
                    string_start -= 2
                if string_start > section_start:
                    string_start += 3  # last step +1 to skip \x00
                elif string_start < section_start:
                    # The leftmost string in the section, so don't step past section_start
                    string_start = section_start
            else:
                string_start = section_start  # The leftmost string in the block

            # Now that we have a string_start and string_end, we can carve the string
            new_decoded_string = decoded_string.decoded_data[string_start:string_end]
            if new_decoded_string:
                # Since we're using \x00 as the delimiter, we need to add \x00 to the end
                new_decoded_string += b"\x00"
                new_string = copy.copy(decoded_string)
                new_string.encoded_data = new_decoded_string
                new_string.decoded_data = new_decoded_string
                new_string.offset = string_start
                results.append(new_string)

            if string_start == section_start:
                break
            else:
                # -1 to skip \x00 for searching (which necessitates adding it back above)
                string_end = string_start - 1

        # We've processed a full section, advance to the next section
        if section_end == len(decoded_string.decoded_data):
            break
        else:
            section_start = section_end + 2
            # Skip blocks of \x00
            while (
                section_start < len(decoded_string.encoded_data)
                and decoded_string.decoded_data[section_start] == 0
            ):
                section_start += 1
            section_end = decoded_string.decoded_data[section_start:].find(b"\x00\x00")
            if section_end == -1:  # The rightmost section in the block
                section_end = len(decoded_string.encoded_data)
            else:
                section_end += section_start
            string_end = section_end

    results.sort(key=lambda decoded_string: decoded_string.start_ea)

    if identify:
        for new_string in results:
            new_string.define()

    return results


def find_encoded_strings_inline(matches, Tracer, **kwargs):
    """
    For each yara match, attempt to find encoded strings.

    Input:
        matches - A list of yara matches (ea, identifier)
        Tracer - A pointer to an implementation of StringTracer.
        **kwargs - kwargs to be passed to Tracer's constructor.

    Output:
        A list of EncodedStrings.
    """
    encoded_strings = []
    for ea, identifier in matches:
        try:
            tracer = Tracer(ea, identifier, **kwargs)
            if tracer.search():
                encoded_strings.extend(tracer.encoded_strings)
            else:
                logger.warning("Failed to find strings at 0x%X" % ea)
        except AttributeError:
            logger.warning("Error tracing at 0x%X" % ea)
    return encoded_strings


def find_encoded_strings(funcs, Tracer, **kwargs):
    """
    Description:
        For each ref for each function, attempt to find encoded strings.

    Input:
        funcs - A list of functions. Must have an xrefs_to field.
        Tracer - A pointer to an implementation of StringTracer.
        **kwargs - kwargs to be passed to Tracer's constructor.

    Output:
        A list of EncodedStrings.
    """
    encoded_strings = []
    for func in funcs:
        for ref in func.xrefs_to:
            if idc.get_segm_name(ref) == ".pdata":
                logger.info(
                    "Segment .pdata for ref 0x%08x is not a relevant code segment and will be skipped"
                    % ref
                )
            else:
                # NOTE: Setting errors to info because it is common and will spam our console.
                try:
                    tracer = Tracer(ref, func.identifier, **kwargs)
                    if tracer.search():
                        encoded_strings.extend(tracer.encoded_strings)
                    else:
                        logger.info("Failed to find strings at 0x%X" % ref)
                except AttributeError as e:
                    # Only catch AttributeErrors resulting from a function not existing. All other AttributeErrors
                    # are actual errors and should go uncaught.
                    #
                    # TODO: Create a separate exception class for a function not existing, so that we don't have to do
                    #  this kind of error message checking.
                    if str(e).startswith("No function at 0x"):  #
                        logger.info(
                            "No function exists at 0x%X. Create a function at this location to obtain strings."
                            % ref
                        )
                    else:
                        raise e
    return encoded_strings


def decode_strings(encoded_strings, decode):
    """
    Description:
        For each encoded_string entry in encoded_strings, decode the data using the provided
        decode function.

    Input:
        encoded_strings - A list of EncodedStrings.
        decode - A pointer to a function that handles decoding.

    Output:
        A list of successfully decoded EncodedStrings.
    """
    # NOTE: Setting errors to info because it is common and will spam our console.
    decoded_strings = []
    for encoded_string in encoded_strings:
        if not encoded_string.encoded_data:
            logger.info("Unable to find string {!r}".format(encoded_string))
            continue
        encoded_string.decoded_data = decode(encoded_string)
        if isinstance(encoded_string.decoded_data, bytearray):
            encoded_string.decoded_data = bytes(encoded_string.decoded_data)
        if encoded_string.decoded_data:  # Allow decoders to abort/fail quietly
            decoded_strings.append(encoded_string)
        else:
            logger.info("Failed to decode string: {!r}".format(encoded_string))
    return decoded_strings


INPUT_FILE_PATH = idc.get_input_file_path()


def find_input_file():
    """
    Description:
        Check whether or not IDA knows where the original file used to create the IDB is.
        If IDA doesn't know, check the IDA's directory for the file.

    Output:
        Returns True if the input file was located, False if it was not.
    """
    global INPUT_FILE_PATH
    ida_path = INPUT_FILE_PATH
    if not os.path.exists(ida_path):
        # If IDA does not know, check if the (correct) file is sitting next to the IDB.
        local_path = os.path.join(idautils.GetIdbDir(), idc.get_root_filename())
        if (
            os.path.exists(local_path)
            and hashlib.md5(open(local_path, "rb").read()).hexdigest().upper()
            == idc.retrieve_input_file_md5()
        ):
            INPUT_FILE_PATH = local_path
            logger.debug("Guessed the input file path: " + INPUT_FILE_PATH)
            logger.debug("IDA thought it was:          " + ida_path)
            return True
        else:
            return False
    else:
        return True


def string_decoder_main(
    yara_rule,
    Tracer,
    decode,
    patch=True,
    func_name="string_decode_function",
    inline=False,
):
    """
    Description:
        If you are going to use this file's workflow as is, this is the entry point. This supports the majority
        of decode function xref based decoding algorithms (the 'older' way) and inline decoding algorithms
        (the 'newer' way).

    Input:
        First, you'll need to provide a YARA rule. The rule can either be passed as a string or in a file.
        If it is in a file, be sure to set is_yara_file = True. Also note that when YARA finds the decoder
        function, it will attempt to rename it:
            func_name = '' will reset the name to the default name it had when IDA built the IDB
            func_name = None will prevent this script from renaming it

        Second, you need to implement the StringTracer class above to handle your malware's particular
        way of loading the encrypted string. Pass a pointer to this class as the second parameter.

        Third, you need to implement the decoding algorithm and pass a pointer to the entry function
        as the third parameter.

        It is also important to have the original file in the correct location for YARA rule application.
        The IDB contains the path of original file from when the IDB was created, but if the file isn't there,
        this script will check the IDB's current directory for a file with the same name and same md5.

        This script will also patch the encoded string with the decoded one unless patch = False.

    Throws:
        AttributeError - see StringTracer, SuperFunc_t
        OSError - The provided YARA file could not be opened.
        RuntimeError - see _yara_find_decode_functions
        TypeError - The provided Tracer does not extend StringTracer.
        ValueError - Could not find the file used to create the IDB.
                   - see EncodedString.get_bytes
    """
    global ENCODED_STRINGS
    global DECODED_STRINGS

    # We could just check if there is a 'search' method, but handle it this way to enforce conventions.
    if not issubclass(Tracer, StringTracer):
        logger.error("Tracer does not extend StringTracer!")
        return

    # Check that IDA actually knows where the original input file is.
    if not find_input_file():
        logger.error(
            "Unable to locate the file used to create the IDB: " + INPUT_FILE_PATH
        )
        return

    # Do the decoding.
    try:
        if inline:
            matches = yara.match_strings(yara_rule)
            ENCODED_STRINGS = find_encoded_strings_inline(matches, Tracer)
        else:
            decode_functions = yara_find_decode_functions(yara_rule, func_name)
            ENCODED_STRINGS = find_encoded_strings(decode_functions, Tracer)
        ENCODED_STRINGS = decode_strings(ENCODED_STRINGS, decode)
        ENCODED_STRINGS = sorted(set(ENCODED_STRINGS))
        for string in ENCODED_STRINGS:
            string.publish(patch=patch)
        return ENCODED_STRINGS
    except RuntimeError as e:
        logger.error(
            "The provided YARA rule failed to match. No strings can be decrypted for this YARA rule."
        )
        return


idc.auto_wait()  # Force wait on import just to be sure



def registrar(registry, name="entry"):
    """
    Creates and returns a register function that can be used as a decorator
    for registering functions into the given registry dictionary.

    :param registry: Dictionary to add entry registrations to.
    :param name: Name to give to each entry. "entry" is used by default.

    :returns: A register() decorator that can be used to fill in the given registry dictionary.
    """

    def register_func(entry_name_or_func):
        """
        Registers an entry for the CPU emulator.
        """
        if callable(entry_name_or_func):
            # If function, that means no argument was passed in and we should register using the function name.
            func = entry_name_or_func
            entry_name = func.__name__.lower()
            registry[entry_name] = func
            return func

        # Otherwise, register with user provided name
        entry_name = entry_name_or_func
        if entry_name in registry:
            raise ValueError("Duplicate {} name: {}".format(name, entry_name))

        def _wrapper(func):
            # Register function as entry.
            registry[entry_name.lower()] = func
            return func  # Must return function afterwards.

        return _wrapper

    return register_func



"""
Description: Sample decoder
Author: DC3
"""

import kordesii
from kordesii.utils import EncodedString
from kordesii.utils import ida_re
from kordesii.utils import function_tracing

logger = kordesii.get_logger()
emulator = function_tracing.Emulator()


def xor_decrypt(key, enc_data):
    return bytes((x ^ key) for x in enc_data)


def find_strings():
    """
    Extracts and publishes EncodedString objects for the parameters following xor encryption function:

        void encrypt(char *s, char key)
        {
            while (*s)
                *s++ ^= key;
        }
    """
    for encrypt_func in ida_re.find_functions(br"\x8b\x45\x08\x0f\xbe\x08"):
        logger.info("Found XOR encrypt function at: 0x%x", encrypt_func.start_ea)
        for call_ea in encrypt_func.calls_to:
            logger.debug("Tracing 0x%08x", call_ea)
            # Extract arguments for call to xor function.
            context, args = emulator.get_function_args(call_ea)
            enc_str_ptr, key = args
            encoded_string = EncodedString(enc_str_ptr, string_reference=call_ea, key=key)
            encoded_string.decoded_data = xor_decrypt(key, encoded_string.encoded_data)
            encoded_string.publish(rename=True, patch=False)


@kordesii.decoder_entry
def main():
    """
    Finds xor encrypted strings then decrypts and outputs them.
    """
    find_strings()



"""
Interfaces for higher level elements such as open files or registry keys.
"""
import ntpath
import logging
from copy import deepcopy
from typing import Iterable, Type, TypeVar

from .actions import *


logger = logging.getLogger(__name__)
T = TypeVar("T")


class ObjectMap(object):
    """
    Interface for obtaining high-level representations of objects encountered during emulation.

    NOTE: Generated objects are immutable and are not bound to the current cpu context after generation.
        (This is different to other item types like Variable, FunctionSignature, Operand, etc.)
    """

    # The bounds on handles that can be used for Objects
    MIN_HANDLE = 0x80
    MAX_HANDLE = 0xFFFFFFFF - 1

    def __init__(self, cpu_context):
        self._cpu_context = cpu_context
        self._next_handle = self.MIN_HANDLE

    def __repr__(self):
        objs = "\n\t".join([repr(obj) for obj in self])
        return f"<ObjectMap : \n\t{objs}\n>"

    def __deepcopy__(self, memo):
        """
        Custom implementation of deepcopy to improve efficiency.
        """
        copy = ObjectMap(deepcopy(self._cpu_context, memo))
        copy._next_handle = self._next_handle
        return copy

    # TODO: Look into caching this if the actions haven't changed.
    #   Perhaps cache within ActionList instead?
    def __getitem__(self, handle: int) -> "Object":
        """Gets an object by handle"""
        if handle not in self:
            raise KeyError(f"Object with handle: {hex(handle)} does not exist.")

        actions = [
            action
            for action in self._cpu_context.actions
            if getattr(action, "handle", None) == handle
        ]
        # Determine the type of object to yield by peeking at the first action.
        # Default to generic Object.
        if actions:
            for object_class in Object.__subclasses__():
                if isinstance(actions[0], object_class.action_types):
                    return object_class(handle, actions)

        # Default to generic Object
        return Object(handle, actions)

    def get(self, handle: int, default=None) -> "Object":
        """Gets an object by handle"""
        try:
            return self[handle]
        except KeyError:
            return default

    def __len__(self) -> int:
        return self._next_handle - self.MIN_HANDLE

    def __contains__(self, handle: int) -> bool:
        return self.MIN_HANDLE <= handle < self._next_handle

    def __bool__(self):
        return self._next_handle > self.MIN_HANDLE

    def _iter(self, reverse=False) -> Iterable["Object"]:
        """
        Yields all known objects based on current set of actions in the cpu context.
        """
        for handle in reversed(self.handles) if reverse else self.handles:
            yield self[handle]

    def __iter__(self):
        """
        Yields all known objects based on current set of actions in the cpu context.
        Iterates least recently used to most.
        """
        yield from self._iter()

    def __reversed__(self):
        """
        Yields all known objects based on current set of actions in cpu context.
        Iterates most recently used to least.
        """
        yield from self._iter(reverse=True)

    @property
    def handles(self) -> List[int]:
        """
        List of allocated handles.
        """
        return list(range(self.MIN_HANDLE, self._next_handle))

    def alloc(self) -> int:
        """
        Allocates and returns the next available handle address.
        """
        if self._next_handle > self.MAX_HANDLE:
            raise ValueError("Too many handles created.")
        handle = self._next_handle
        self._next_handle += 1
        return handle

    def get_or_alloc(self, obj_type: Type["Object"], **query) -> int:
        """
        Returns the handle of a known object or a new handle if an object
        of type `obj_type` containing the attribute(s) equivalent to those
        found in `query` does not exist.
        """
        # First grab the most recent object containing the handle.
        # Reverse to ensure we get the most recent one.
        for obj in self.query(obj_type, reverse=True, **query):
            return obj.handle
        return self.alloc()

    def query(self, obj_type: Type[T], reverse=False, **conditions) -> Iterable[T]:
        """
        Returns the handle of a known object
        :param obj_type: Type of object to query for.
        :param reverse: Whether to produce objects from most recently used to least.
        :param conditions: Attributes to look for.
        :return:
        """
        for obj in (reversed(self) if reverse else self):
            if isinstance(obj, obj_type) \
                    and all(getattr(obj, attr_name) == value for attr_name, value in conditions.items()):
                yield obj

    def at(self, ip: int) -> List["Object"]:
        """
        Retrieves the objects referenced at the given instruction address.

        :param ip: Instruction address to get pointers from.
        :return: List of Object objects that were found within the given instruction.

        :raises ValueError: If instruction has not been executed yet.
        """
        if ip not in self._cpu_context.executed_instructions:
            raise ValueError(
                f"Unable to get objects. Instruction at 0x{ip:0x} has not been executed."
            )
        return [obj for obj in self if ip in obj.references]


class Object(object):
    """
    Represents an high level instantiated object during emulation.
    """

    # The type of Actions that builds the object.
    # TODO: Create a __new__() to validate action types are unique per class.
    action_types = tuple()

    def __init__(self, handle: int, actions: List[Action]):
        # The list of actions that are relevant to this object.
        self.handle = handle
        self.actions = actions

    @property
    def references(self) -> List[int]:
        """The address locations where this object has been encountered."""
        return [action.ip for action in self.actions]


class File(Object):
    """
    Stores information for opened files for a specific CPU context state.
    """

    action_types = (
        FileCreated, FileOpened, FileTruncated, FileDeleted, FileMoved, FileWritten, FileClosed
    )
    file_creation_types = (FileCreated, FileOpened, FileTruncated)

    def __repr__(self):
        data = self.data
        data_str = repr(data[:10])
        if len(data) > 10:
            data_str = data_str[:-1] + "..." + data_str[-1]
        return (
            f"<File 0x{self.handle:08X}"
            f" : path = {self.path}"
            f" : mode = {self.mode}"
            f" : size = {len(data)}"
            f" : data = {data_str}"
            f">"
        )

    @property
    def data(self):
        """The data written to the file."""
        return b"".join([action.data for action in self.actions if isinstance(action, FileWritten)])

    @property
    def path(self) -> Optional[str]:
        """The path of the file."""
        for action in reversed(self.actions):
            if isinstance(action, self.file_creation_types + (FileDeleted,)):
                return action.path
            elif isinstance(action, FileMoved):
                return action.new_path

    @property
    def name(self) -> Optional[str]:
        """The base name of the file."""
        if self.path:
            return ntpath.basename(self.path)

    @property
    def history(self) -> List[str]:
        """List of previous file paths."""
        history = []
        for action in self.actions:
            if isinstance(action, FileMoved):
                history.append(action.old_path)
        return history

    @property
    def mode(self) -> Optional[str]:
        """The mode the file was last opened with."""
        for action in reversed(self.actions):
            if isinstance(action, self.file_creation_types):
                return action.mode

    @property
    def closed(self) -> Optional[bool]:
        """
        Whether or not the file has been closed.
        If None, this information is unknown.
        """
        for action in reversed(self.actions):
            if isinstance(action, FileClosed):
                return True
            elif isinstance(action, self.file_creation_types):
                return False

    @property
    def deleted(self) -> Optional[bool]:
        """
        Whether or not the file has been deleted.
        If None, this information is unknown.
        """
        for action in reversed(self.actions):
            if isinstance(action, FileDeleted):
                return True
            elif isinstance(action, self.file_creation_types):
                return False


class RegKey(Object):
    """
    Stores information for opened registry keys for a specific CPU context state.
    """

    action_types = (
        RegKeyOpened, RegKeyDeleted, RegKeyValueDeleted, RegKeyValueSet
    )

    def __repr__(self):
        return (
            f"<RegKey 0x{self.handle:08X}"
            f" : root_key = {self.root_key}"
            f" : sub_key = {self.sub_key}"
            f">"
        )

    @property
    def root_key(self) -> Optional[str]:
        """The root key of the registry key."""
        for action in reversed(self.actions):
            if isinstance(action, RegKeyOpened):
                return action.root_key

    @property
    def sub_key(self) -> Optional[str]:
        """The sub key of the registry key."""
        for action in reversed(self.actions):
            if isinstance(action, RegKeyOpened):
                return action.sub_key

    @property
    def path(self) -> Optional[str]:
        """The full path of the registry key."""
        if self.root_key and self.sub_key:
            return "\\".join([self.root_key, self.sub_key])


class Service(Object):
    """
    Stores information pertaining to a service
    """

    action_types = (
        ServiceCreated, ServiceOpened, ServiceDeleted, ServiceDescriptionChanged
    )

    def __repr__(self):
        return (
            f"<Service 0x{self.handle:08X}"
            f" : name = {self.name}"
            f" : description = {self.description}"
            f">"
        )

    @property
    def name(self) -> Optional[str]:
        """The name of the service."""
        for action in reversed(self.actions):
            if isinstance(action, ServiceCreated) or isinstance(action, ServiceOpened):
                return action.name

    @property
    def description(self) -> Optional[str]:
        """The description of the service"""
        for action in reversed(self.actions):
            if isinstance(action, ServiceDescriptionChanged):
                return action.description



"""
Utilities for running YARA within IDA.

This utility extends and overwrites the existing YARA API to work correctly within IDA.

usage::
    from kordesii.utils import yara

    rule = yara.compile(source=rule_text)
    matches = rule.match()  # No parameter needed for match to run on entire input file.

    # Can also be used to run on segments.
    matches = rule.match(segment='_text')

    # We can also just look for matching strings.
    for offset, identifer in rule.match_strings():
        # ...
    for offset, identifer in rule.match_strings(segment='_text'):
        # ...
"""

import logging
from typing import Iterable, List, Tuple

import idc
import idaapi
import idc
import yara
from yara import *

from kordesii.utils import segments
from kordesii.utils.functions import Function


logger = logging.getLogger(__name__)

READ_LENGTH = 10485760  # 10 MB


class Match(object):
    """
    Patches yara.Match to provide convert string offsets to virtual addresses.

    NOTE: We can't inherit yara.Match because they don't expose that class.

    :param yara.Match match_object: Original match object created by YARA
    :param int offset: Optional offset to offset string offsets by
    :param bool input_offset: Whether string offsets will be the file offset
        and should be converted.
    """

    def __init__(self, match_object, offset=None, file_offset=False):
        self._match = match_object
        self._offset = offset
        self._file_offset = file_offset
        self._strings = None

    def __getattr__(self, item):
        return getattr(self._match, item)

    def __str__(self):
        return str(self._match)

    def __repr__(self):
        return repr(self._match)

    @property
    def strings(self):
        # Before returning strings, fixup the offsets to be virtual addresses.
        if self._strings is None:
            self._strings = []
            for offset, identifier, data in self._match.strings:
                if self._offset is not None:
                    offset += self._offset
                if self._file_offset:
                    offset = idaapi.get_fileregion_ea(offset)
                self._strings.append((idc.get_item_head(offset), identifier, data))
        return self._strings


class Rules(object):
    """
    Patches yara.Rules to use our patched Match object when match() is called.

    NOTE: We can't inherit yara.Rules because they don't expose that class.
    """

    def __init__(self, rules_object):
        self._rules = rules_object
        self._infos = None

    def __getattr__(self, item):
        return getattr(self._rules, item)

    def _extract_info(self):
        """
        Retrieve information about the rule by performing a kludgy dance with callbacks.
        YARA should allow an easier way to give this information!
        """
        if self._infos is None:
            # YARA doesn't provide any easy way to get rule info, so we are going to have
            # to fake a match to get the info dictionary.
            self._infos = []

            def _callback(info):
                self._infos.append(info)
                return yara.CALLBACK_CONTINUE

            self._rules.match(data=b"", callback=_callback, which_callbacks=yara.CALLBACK_NON_MATCHES)
        return self._infos

    @property
    def names(self) -> List[str]:
        """Returns names of all the rules contained within."""
        infos = self._extract_info()
        return [info["rule"] for info in infos]

    def match(self, *args, **kwargs) -> List[Match]:
        """
        Patched to use our patched Match() object and allow for automatically running
        on IDB input file.

        Besides the default yara parameters, this implementation also includes:
            :param bool input_offset: Whether to apply input file offset to string offsets.
            :param int offset: Optional offset to offset string offsets by.
            :param str|int segment: Name or EA of segment to match to.
        """
        input_offset = kwargs.pop("input_offset", False)
        offset = kwargs.pop("offset", None)
        segment = kwargs.pop("segment", None)

        # Run on segment.
        if segment:
            kwargs["data"] = segments.get_bytes(segment)
            offset = offset or segments.get_start(segment)
        # Run on input file.
        elif not (args or kwargs):
            args = (idc.get_input_file_path(),)
            input_offset = True

        return [Match(match, offset=offset, file_offset=input_offset) for match in self._rules.match(*args, **kwargs)]

    def match_strings(self, *args, **kwargs) -> List[Tuple[int, str]]:
        """
        Runs match() but then returns tuples containing matched strings instead of Match objects.

        (This replicates the original legacy _YARA_MATCHES output)

        :returns: Tuple containing: (offset, identifier)
        """
        matched_strings = []
        for match in self.match(*args, **kwargs):
            for offset, identifier, _ in match.strings:
                matched_strings.append((offset, identifier))
        return matched_strings


def compile(*args, **kwargs):
    """Wraps compiled rule in our patched Rules object."""
    return Rules(yara.compile(*args, **kwargs))


def load(*args, **kwargs):
    """Wraps loaded rule in our patched Rules object."""
    return Rules(yara.load(*args, **kwargs))


# Convenience functions ==============


def match(rule_text: str, *args, **kwargs) -> List[Match]:
    """Returns list of Match objects"""
    rule = compile(source=rule_text)
    return rule.match(*args, **kwargs)


def match_strings(rule_text: str, *args, **kwargs) -> List[Tuple[int, str]]:
    """Returns list of (offset, string identifier)"""
    rule = compile(source=rule_text)
    return rule.match_strings(*args, **kwargs)


def find_functions(rule_text: str, *args, **kwargs) -> Iterable[Function]:
    """
    Iterates functions that match the given rule text.
    """
    rule = compile(source=rule_text)

    cache = set()
    for match in rule.match(*args, **kwargs):
        for ea, identifier, _ in match.strings:
            try:
                func = Function(ea, identifier)
            except AttributeError:
                continue

            if func not in cache:
                cache.add(func)
                yield func

# ====================================



"""
Utilty functions for the ARM architecture.
"""



# region Shift Operations


def lsl(value, count, width=32):
    """Logical Shift Left"""
    value <<= count
    carry = (value & (1 << width)) >> width
    value &= (1 << width) - 1
    return carry, value


def lsr(value, count, width=32):
    """Logical Shift Right"""
    count %= width
    value &= (1 << width) - 1
     # First shift 1 to the left to leave room for the carry.
    value <<= 1
    value >>= count
    carry = value & 1
    value >>= 1
    return carry, value


def asr(value, count, width=32):
    """Arithmetic Shift Right (assuming unsigned input)"""
    count %= width
    value &= (1 << width) - 1
    msb = value >> (width - 1)
    value -= (1 << width) * msb  # Convert to signed.
    # First shift 1 to the left to leave room for the carry.
    value <<= 1
    value >>= count
    carry = value & 1
    value >>= 1
    value &= (1 << width) - 1  # Convert back to unsigned.
    return carry, value


def ror(value, count, width=32):
    """
    Rotate Right

    >>> ror(0xff, 4) == (1, 0xf000000f)
    True
    >>> ror(0xff00, 8, width=16) == (1, 0xff)
    True
    >>> ror(0x1, 2) == (0, 0x40000000)
    True
    """
    count %= width
    value &= (1 << width) - 1
    if not count:
        return 0, value
    # First rotate value count - 1 times
    value = (value >> (count - 1)) | (value << (width - (count - 1)))
    # Then pull out the carry before rotating one more time.
    carry = value & 1
    value = (value >> 1) | (value << (width - 1))
    value &= (1 << width) - 1
    return carry, value


def rrx(carry, value, count, width=32):
    """
    Rotate Right Extended
    (original carry must also be passed in)

    >>> rrx(0, 0xff, 4) == (1, 0xe000000f)
    True
    >>> rrx(1, 0xff, 4) == (1, 0xf000000f)
    True
    """
    count %= width
    value &= (1 << width) - 1
    # First shift 1 to the left to leave room for the carry.
    value <<= 1
    # Stick in the original carry.
    value |= carry & 1
    value = (value >> count) | (value << ((width + 1) - count))
    carry = value & 1
    value >>= 1
    value &= (1 << width) - 1
    return carry, value


# endregion



"""
Test case support for DC3-Kordesii. Decoder output is stored in a json file per decoder. To run test cases,
decoder is re-run and compared to previous results.
"""

from io import open
from timeit import default_timer
import difflib
import json
import logging
import multiprocessing as mp
import os
import pathlib
import textwrap

import kordesii.reporter
from kordesii import registry
from kordesii import logutil

logger = logging.getLogger(__name__)


FIELD_LAST_UPDATE = "last_update"

# Constants
INPUT_FILE_PATH = "input_file"
FILE_EXTENSION = ".json"
DECODER = "decoder"
RESULTS = "results"
PASSED = "passed"
ERRORS = "errors"
DEBUG = "debug"
IDA_LOG = "ida_log"
RUN_TIME = "run_time"


DEFAULT_EXCLUDE_FIELDS = (
    kordesii.reporter.FIELD_DEBUG,
    "idb",  # Legacy field that should no longer exists. TODO: remove in future version.
    FIELD_LAST_UPDATE,
)

DEFAULT_INCLUDE_FIELDS = (kordesii.reporter.FIELD_STRINGS, kordesii.reporter.FIELD_FILES)


def multiproc_test_wrapper(args):
    """Wrapper function for running tests in multiple processes."""
    test_case = args[0]
    try:
        return test_case.run(*args[1:])
    except KeyboardInterrupt:
        return


def multiproc_initializer(log_level, log_port, decoder_sources, default_source):
    """Multiproc initializer used to pass along log level/port and registry information."""
    registry._sources = decoder_sources
    registry._default_source = default_source
    logging.root.setLevel(log_level)
    if log_port:
        logutil._started_listener = True
        logutil.listen_port = log_port


class Tester(object):
    """
    DC3-Kordesii Tester class
    """

    def __init__(
        self,
        reporter,
        results_dir=None,
        decoder_names=None,
        nprocs=None,
        field_names=None,
        ignore_field_names=DEFAULT_EXCLUDE_FIELDS,
        malware_repo=None,
    ):
        """
        Run tests and compare produced results to expected results.

        :param kordesii.Reporter reporter: Kordesii reporter object
        :param str results_dir: Directory of json test cases
                (defaults to dynamically pulling from a "tests" folder within parser's directory.)
        :param [str] decoder_names:
                A list of decoder names to run tests for. If the list is empty (default),
                then test cases for all parsers will be run.
        :param [str] field_names:
                A restricted list of fields (metadata key values) that should be compared
                during testing. If the list is empty (default), then all fields, except those in
                ignore_field_names will be compared.
        :param int nprocs: Number of processes to use. (defaults to (3*num_cores)/4)
        """
        self.reporter = reporter
        self.results_dir = results_dir
        self.decoder_names = decoder_names or [None]
        self.field_names = field_names or []
        self.ignore_field_names = ignore_field_names
        self.malware_repo = malware_repo
        self._test_cases = None
        self._results = []  # Cached results.
        self._processed = False
        self._nprocs = nprocs or (3 * mp.cpu_count()) // 4

    def __iter__(self):
        return self._iter_results()

    def _iter_results(self):
        # First yield any cached results.
        for result in self._results:
            yield result

        # Run tests in multiprocessing pool (if not already run)
        if not self._processed:
            self._processed = True
            log_level = logging.root.getEffectiveLevel()
            pool = mp.Pool(
                processes=self._nprocs,
                initializer=multiproc_initializer,
                initargs=(log_level, logutil.listen_port, registry._sources, registry._default_source),
            )
            test_iter = pool.imap_unordered(multiproc_test_wrapper, [(test_case,) for test_case in self.test_cases])
            pool.close()

            try:
                for result in test_iter:
                    self._results.append(result)
                    yield result
            except KeyboardInterrupt:
                pool.terminate()
                raise

    @property
    def test_cases(self):
        """Returns test cases."""
        if self._test_cases is not None:
            return self._test_cases

        self._test_cases = []
        for decoder_name in self.decoder_names:
            # We want to iterate decoders in cases decoder_name represents a set of decoders from
            # different sources.
            found = False
            for decoder in kordesii.iter_decoders(decoder_name):
                found = True
                results_file_path = self.get_results_filepath(decoder.full_name)
                if os.path.isfile(results_file_path):
                    for expected_results in self.read_results_file(results_file_path):
                        self._test_cases.append(
                            TestCase(
                                self.reporter,
                                decoder.full_name,
                                expected_results,
                                field_names=self.field_names,
                                ignore_field_names=self.ignore_field_names,
                            )
                        )
                else:
                    logger.warning("Test case file not found: {}".format(results_file_path))

            if not found and decoder_name:
                # Add a failed result if we have an orphan test.
                self._results.append(TestResult(decoder_name=decoder_name, passed=False, errors=["Decoder not found."]))
        return self._test_cases

    @property
    def total(self):
        """Returns total number of results."""
        return len(self._results) + len(self.test_cases)

    def gen_results(self, decoder_name, input_file_path):
        """
        Generate JSON results for the given file using the given decoder name.
        """
        # Read in data so we avoid placing idb files in the malware repo.
        with open(input_file_path, "rb") as f:
            data = f.read()
        self.reporter.run_decoder(decoder_name, data=data, log=True)
        self.reporter.metadata[INPUT_FILE_PATH] = os.path.abspath(input_file_path)
        return self.reporter.metadata

    def _list_test_files(self, results_list):
        """
        Returns a list of the input file paths for the given results_list.
        """
        return [results[INPUT_FILE_PATH] for results in results_list]

    def get_results_filepath(self, decoder_name):
        """
        Get a results file path based on the decoder name provided and the
        previously specified output directory.
        """
        for decoder in kordesii.iter_decoders(decoder_name):
            file_name = decoder.name + FILE_EXTENSION
            # Use hardcoded results dir if requested.
            if self.results_dir:
                return os.path.join(self.results_dir, file_name)

            # Assume there is a "tests" folder within the source path.
            test_dir = os.path.join(decoder.source.path, "tests")
            return os.path.normpath(os.path.join(test_dir, file_name))

        raise ValueError("Invalid parser: {}".format(decoder_name))

    def read_results_file(self, results_file_path):
        """
        Parse the the JSON results file and return the parsed data.
        """
        if not os.path.exists(results_file_path):
            results = []
        else:
            with open(results_file_path, "r", encoding="utf8") as results_file:
                results = json.load(results_file)

        # The results file data is expected to be a list of metadata dictionaries
        if not isinstance(results, list) or not all(isinstance(a, dict) for a in results):
            raise ValueError("Results file is invalid: {}".format(results_file_path))

        # Replace {MALAWARE_REPO} with full input file.
        for result in results:
            # Add results_file_path for relative paths.
            # NOTE: Using PureWindowsPath to help convert a Windows path using \
            #   into / path.
            #   This helps in-case the test case was originally made with a Windows machine
            #   but is being tested on Linux.
            input_file_path = pathlib.PureWindowsPath(result[INPUT_FILE_PATH]).as_posix()
            # expand environment variables
            input_file_path = os.path.expandvars(input_file_path)
            # resolve variables
            if self.malware_repo:
                input_file_path = input_file_path.format(MALWARE_REPO=self.malware_repo)

            input_file_path = os.path.join(os.path.dirname(results_file_path), input_file_path)
            input_file_path = os.path.abspath(input_file_path)
            result[INPUT_FILE_PATH] = input_file_path

        return results

    def write_results_file(self, results_list, file_path):
        """
        Saves the JSON results list to the given file path.
        :param list[dict] results_list: JSON results list to save
        :param str file_path: Path to save the results JSON file.
        """
        # Replace references to the malware repo with a variable.
        if self.malware_repo:
            for results in results_list:
                input_file_path = results[INPUT_FILE_PATH]
                if input_file_path.startswith(self.malware_repo):
                    input_file_path = "{MALWARE_REPO}" + input_file_path[len(self.malware_repo) :]
                results[INPUT_FILE_PATH] = input_file_path

        # Alphabetize decoded strings
        for results in results_list:
            if kordesii.reporter.FIELD_STRINGS in results:
                results[kordesii.reporter.FIELD_STRINGS] = sorted(results[kordesii.reporter.FIELD_STRINGS])

        # Write updated data to results file
        # NOTE: We need to use dumps instead of dump to avoid TypeError.
        with open(file_path, "w", encoding="utf8") as results_file:
            results_file.write(str(json.dumps(results_list, indent=4, sort_keys=True)))

    def update_tests(self, force=False):
        """
        Updates existing test cases by rerunning parsers.

        :param bool force: Whether to force adding the test case even if errors are encountered
        """
        orig_level = logging.root.level
        logging.root.setLevel(logging.INFO)  # Force info level logs so test cases stay consistent.
        try:
            for decoder_name in self.decoder_names:
                results_file_path = self.get_results_filepath(decoder_name)
                if not os.path.isfile(results_file_path):
                    logger.warning("No test case file found for decoder: {}")
                    continue
                results_list = self.read_results_file(results_file_path)
                for index, file_path in enumerate(self._list_test_files(results_list)):
                    new_results = self.gen_results(decoder_name, file_path)
                    if not new_results:
                        logger.warning(
                            "Empty results for {} in {}, not updating.".format(file_path, results_file_path)
                        )
                        continue
                    if self.reporter.errors and not force:
                        logger.warning("Results for {} has errors, not updating.".format(file_path))
                        continue

                    logger.info("Updating results for {} in {}".format(file_path, results_file_path))
                    results_list[index] = new_results
                self.write_results_file(results_list, results_file_path)
        finally:
            logging.root.setLevel(orig_level)

    def add_test(self, file_path, force=False):
        """
        Adds test case for given file path.

        :param str file_path: Path to input file to add.
        :param bool force: Whether to force adding the test case even if errors are encountered
        """
        orig_level = logging.root.level
        logging.root.setLevel(logging.INFO)  # Force info level logs so test cases stay consistent.
        try:
            for decoder_name in self.decoder_names:
                results_file_path = self.get_results_filepath(decoder_name)
                results_list = self.read_results_file(results_file_path)
                input_files = self._list_test_files(results_list)

                if file_path in input_files:
                    logger.warning("Test case for {} already exists in {}".format(file_path, results_file_path))
                    continue

                new_results = self.gen_results(decoder_name, file_path)
                if not new_results:
                    logger.warning("Empty results for {} in {}, not adding.".format(file_path, results_file_path))
                    continue
                if self.reporter.errors and not force:
                    logger.warning("Results for {} has errors, not adding.".format(file_path))
                    continue

                logger.info("Adding results for {} in {}".format(file_path, results_file_path))
                results_list.append(new_results)
                self.write_results_file(results_list, results_file_path)
        finally:
            logging.root.setLevel(orig_level)

    def remove_test(self, file_path):
        """Removes test case for given file path."""
        for decoder_name in self.decoder_names:
            results_file_path = self.get_results_filepath(decoder_name)
            results_list = []
            removed = False
            for results in self.read_results_file(results_file_path):
                if results[INPUT_FILE_PATH] == file_path:
                    logger.info("Removed results for {} in {}".format(file_path, results_file_path))
                    removed = True
                else:
                    results_list.append(results)

            if removed:
                self.write_results_file(results_list, results_file_path)


class TestCase(object):
    def __init__(self, reporter, decoder_name, expected_results, field_names=None, ignore_field_names=None):
        self._reporter = reporter
        self.decoder_name = decoder_name
        self.expected_results = expected_results
        self.input_file_path = expected_results["input_file"]
        self.filename = os.path.basename(self.input_file_path)
        self._field_names = field_names or []
        if ignore_field_names is None:
            ignore_field_names = DEFAULT_EXCLUDE_FIELDS
        self._ignore_field_names = ignore_field_names

    def run(self):
        """Run test case."""
        start_time = default_timer()

        # Read in data so we avoid placing idb files in the malware repo.
        with open(self.input_file_path, "rb") as f:
            data = f.read()
        self._reporter.run_decoder(self.decoder_name, data=data, log=True)
        self._reporter.metadata[INPUT_FILE_PATH] = self.input_file_path
        results = self._reporter.metadata

        comparer_results = self._compare_results(self.expected_results, results)
        passed = all(comparer.passed for comparer in comparer_results)

        done_time = default_timer()
        run_time = done_time - start_time

        return TestResult(
            decoder_name=self.decoder_name,
            passed=passed,
            input_file_path=self.input_file_path,
            errors=self._reporter.errors,
            debug=self._reporter.metadata.get("debug", None),
            ida_log=self._reporter.ida_log,
            results=comparer_results,
            run_time=run_time,
        )

    def _compare_results(self, results_a, results_b):
        """
        Compare two result sets. If the field names list is not empty,
        then only the fields (metadata key values) in the list will be compared.
        ignore_field_names fields are not compared unless included in field_names.
        """
        results = []

        # Cursory check to remove FILE_INPUT_PATH key from results since it is
        # a custom added field for test cases
        if INPUT_FILE_PATH in results_a:
            results_a = dict(results_a)
            del results_a[INPUT_FILE_PATH]
        if INPUT_FILE_PATH in results_b:
            results_b = dict(results_b)
            del results_b[INPUT_FILE_PATH]

        # Begin comparing results
        if self._field_names:
            for field_name in self._field_names:
                try:
                    comparer = self._compare_results_field(results_a, results_b, field_name)
                except Exception as e:
                    comparer = ResultComparison(field_name)
                    logger.exception(e)
                results.append(comparer)

        else:
            for ignore_field in self._ignore_field_names:
                results_a.pop(ignore_field, None)
                results_b.pop(ignore_field, None)
            all_field_names = set(results_a.keys()) | set(results_b.keys())
            if "other_data" not in set(results_a.keys()) & set(results_b.keys()):
                all_field_names -= {"other_data"}
            for field_name in all_field_names:
                try:
                    comparer = self._compare_results_field(results_a, results_b, field_name)
                except Exception as e:
                    comparer = ResultComparison(field_name)
                    logger.exception(e)
                results.append(comparer)

        return results

    def _compare_results_field(self, results_a, results_b, field_name):
        """
        Compare the values for a single results field in the two passed in results.
        """
        assert isinstance(results_a, dict) and isinstance(results_b, dict)
        comparer = ResultComparison(field_name)

        # Confirm key is found in at least one of the passed in result sets
        if field_name not in results_a and field_name not in results_b:
            return comparer

        # Compare results and return result
        result_a = results_a.get(field_name, [])
        result_b = results_b.get(field_name, [])

        # Compare results and return result
        comparer.compare(result_a, result_b)
        return comparer


class TestResult(object):
    def __init__(
        self,
        decoder_name,
        passed,
        input_file_path=None,
        errors=None,
        debug=None,
        ida_log=None,
        results=None,
        run_time=None,
    ):
        self.decoder_name = decoder_name
        self.input_file_path = input_file_path or "N/A"
        self.filename = os.path.basename(input_file_path) if input_file_path else "N/A"
        self.passed = passed
        self.errors = errors or []
        self.debug = debug or []
        self.ida_log = ida_log or []
        self.results = results or []
        self.run_time = run_time or 0

    def print(self):
        """
        print test result based on provided parameters.
        """
        filtered_output = ""
        filtered_output += "decoder: {}\n".format(self.decoder_name)
        filtered_output += "input_file: {}\n".format(self.input_file_path)
        filtered_output += "passed: {}\n".format(self.passed)
        # Add logs if failed.
        if not self.passed:
            filtered_output += "errors: {}".format("\n" if self.errors else "None\n")
            if self.errors:
                for entry in self.errors:
                    filtered_output += "\t{}\n".format(entry)
            filtered_output += "debug: {}".format("\n" if self.debug else "None\n")
            if self.debug:
                for entry in self.debug:
                    filtered_output += "\t{}\n".format(entry)
            if self.results:
                filtered_output += "results:\n"
                for result in self.results:
                    if not result.passed:
                        filtered_output += "{}\n".format(result)
            if self.errors:
                filtered_output += "ida_log:\n{}\n".format(self.ida_log)
        filtered_output += "\n"

        print(filtered_output.encode("ascii", "backslashreplace").decode())


class ResultComparison(object):
    def __init__(self, field):
        self.field = field
        self.passed = False
        self.missing = []  # Entries found in test case but not new results
        self.unexpected = []  # Entries found in new results but not test case
        self._diff_report = None

    def compare(self, test_case_results, new_results):
        """Compare two result sets and document any differences."""

        if isinstance(test_case_results, str) and isinstance(new_results, str):
            return self._diff(test_case_results, new_results)

        self.missing = []
        self.unexpected = []

        # TODO This could be handled when the JSON test file is read
        # The newly generated results are already unicode-escape'd while
        # the strings from the test file are utf-8 encoded unicode-escape'd strings
        # test_case_results = [tc.encode('unicode-escape').decode('utf8') for tc in test_case_results]
        try:
            test_case_results = [
                tc.encode("utf8").decode("unicode-escape", errors="backslashreplace") for tc in test_case_results
            ]
            new_results = [nr.encode("utf8").decode("unicode-escape", errors="backslashreplace") for nr in new_results]
        except AttributeError:
            pass

        for item in test_case_results:
            if len(item) > 0 and item not in new_results:
                self.missing.append(item)

        for item in new_results:
            if len(item) > 0 and item not in test_case_results:
                self.unexpected.append(item)

        self.passed = not bool(self.missing or self.unexpected)

    def get_report(self, json=False, tabs=1):
        """
        If json parameter is False, get report as a string.
        If json parameter is True, get report as a dictionary.
        """
        if json and not self._diff_report:
            return self.__dict__
        elif json and self._diff_report:
            return {"diff": self._diff_report}
        elif self._diff_report:
            return self._diff_report
        else:
            report = f"{self.field}:\n"
            report += f"\tPassed: {self.passed}\n"
            if self.missing:
                report += "\tMissing:\n"
                for item in self.missing:
                    report += f"\t\t{repr(item)}\n"
            if self.unexpected:
                report += "\tUnexpected:\n"
                for item in self.unexpected:
                    report += f"\t\t{repr(item)}\n"

            return textwrap.indent(report, "\t" * tabs).rstrip()

    def _diff(self, test_case_results: str, new_results: str):
        diff = difflib.context_diff(
            test_case_results.splitlines(True), new_results.splitlines(True), "Test Case", "New Results", n=0
        )

        self._diff_report = "".join(diff)

        self.passed = test_case_results == new_results

    def __str__(self):
        return self.get_report()

    def __repr__(self):
        return self.__str__()


class MyEncoder(json.JSONEncoder):
    def default(self, o):
        return o.__dict__



"""
Utility for running re within IDA.

This utility extends and overwrites the existing re API to work correctly within IDA.
This module works just like the builtin re module, but adjusts offsets to be virtual addresses
and allows for searching specific segments.

usage::
    from kordesii.utils import ida_re


    ptn = ida_re.compile('some pattern')

    for match in ptn.finditer('.text'):
        print('found marker at 0x{:0x}'.format(match.start()))
"""

import re
from typing import Iterable

import idautils
import ida_segment

from kordesii.utils import segments
from kordesii.utils.functions import Function


class Match(object):
    """
    Wraps the SRE_Match object returned by re.
    """

    def __init__(self, match, seg_start):
        self._match = match
        self._start = seg_start

    def __getattr__(self, item):
        """
        Redirects anything that this class doesn't support back to the matchobject class

        :param item:

        :return:
        """
        return getattr(self._match, item, None)

    def start(self, group=None):
        """
        Returns the match object start value with respect to the segment start.

        :param group: optional group to obtain the start of

        :return: virtual start address
        """
        if group:
            _group_start = self._match.start(group)
            if _group_start < 0:  # group exists, but not contributing to match
                return _group_start

            return _group_start + self._start

        return self._match.start() + self._start

    def end(self, group=None):
        """
        Returns the match object end value with respect to the segment start.

        :param group: optional group to obtain the end of

        :return: virtual end address
        """
        if group:
            _group_end = self._match.end(group)
            if _group_end < 0:  # group exists, but not contributing to match
                return _group_end

            return _group_end + self._start

        return self._match.end() + self._start


class Pattern(object):
    """
    Wraps the SRE_Pattern object returned by re.
    """

    def __init__(self, ptn, flags=0):
        if isinstance(ptn, (str, bytes)):
            self._re = re.compile(ptn, flags=flags)
        else:
            self._re = ptn

    @property
    def pattern(self):
        return self._re.pattern

    def _get_segments(self, segname=None):
        """
        Obtain the bytes of the segment specified in segname or all segments as an iterable.

        :param str segname: segment name or None

        :yield: seg_start, seg_bytes
        """
        if segname:
            seg_starts = [ida_segment.get_segm_by_name(segname).start_ea]
        else:
            seg_starts = idautils.Segments()

        for ea in seg_starts:
            yield ea, segments.get_bytes(ea)

    def search(self, segname=None):
        """
        Performs the search functionality on the entire file, searching each segment individually.

        :return: match object modified to match the segment start address
        """
        for seg_start, seg_bytes in self._get_segments(segname):
            match = self._re.search(seg_bytes)
            if match:
                return Match(match, seg_start)
        return None

    def finditer(self, segname=None):
        """
        Performs the finditer functionality on the entire file, searching each segment individually.

        :param segname: Restrict searching to segment with provided name

        :yield: match object
        """
        for seg_start, seg_bytes in self._get_segments(segname):
            for match in self._re.finditer(seg_bytes):
                yield Match(match, seg_start)

    def findall(self, segname=None):
        """
        Performs the findall functionality on the entire file.

        :return: list of match objects
        """
        matches = []
        for _, seg_bytes in self._get_segments(segname):
            matches.extend(self._re.findall(seg_bytes))

        return matches


def compile(pattern, flags=0):
    """Compile a regular expression returning a Pattern object."""
    return Pattern(pattern, flags=flags)


def search(pattern, segname=None, flags=0):
    """Search with regular expression, returning a Match object."""
    return Pattern(pattern, flags=flags).search(segname=segname)


def finditer(pattern, segname=None, flags=0):
    """Iterator of non-overlapping matches."""
    ptn = Pattern(pattern, flags=flags)
    for match in ptn.finditer(segname=segname):
        yield match


def findall(pattern, segname=None, flags=0):
    """Returns a list of non-overlapping matches."""
    return Pattern(pattern, flags=flags).findall(segname=segname)


def find_functions(pattern, flags=0, segname=None) -> Iterable[Function]:
    """
    Uses finditer() to search for functions that contains a match for the given
    regular expression pattern.

    Yields kordesii.utils.Function object for each function.
    """
    if isinstance(pattern, Pattern):
        _regex = pattern
    else:
        _regex = Pattern(pattern, flags)

    cache = set()
    for match in _regex.finditer(segname):
        try:
            func = Function(match.start())
        except AttributeError:
            continue

        if func not in cache:
            cache.add(func)
            yield func



"""
Provides user-facing API for kordesii.utils
"""

# Expose common classes and functions
from .flowchart import Flowchart, Flowchart, BasicBlock
from .functions import Function
from .strings import EncodedString, EncodedStackString, find_string_data
from .utils import *



"""
Provides user-facing API for function_tracing.
"""

# Import functions and classes we want to expose as the API.
from .emulator import Emulator
from .cpu_context import ProcessorContext
# TODO: Remove import for function_tracer in future release.
from .function_tracer import FunctionTracer, TracerCache, get_tracer, hook_tracers, clear_hooks
from .flowchart import FlowChart, Flowchart, BasicBlock
from .objects import File, RegKey, Service
from .actions import *
from .exceptions import *

# Expose constants that may be useful
from .constants import *

# TODO: Can these be generalized for all architectures?
# TODO: Move the STACK_BASE variable into the Emulator and remove RSP_OFFSET and RBP_OFFSET entirely.
from .x86_64 import STACK_BASE, RSP_OFFSET, RBP_OFFSET
from .x86_64 import x86_64ProcessorContext
from .ARM import ARMProcessorContext



"""
Interface for instruction management.
"""
import logging
from copy import deepcopy
from typing import List
import warnings

import ida_frame
import ida_funcs
import ida_idp
import ida_struct
import ida_ua
import idc
import ida_allins

from kordesii.utils.function_tracing import utils
from kordesii.utils.function_tracing.exceptions import FunctionTracingError
from kordesii.utils.function_tracing.operands import Operand

logger = logging.getLogger(__name__)


class Instruction:

    # Cache for keeping track of instructions and their operand indexes.
    _operand_indices = {}

    # Class used for creating new operands.
    _operand_class = Operand

    def __init__(self, cpu_context, ip):
        self.ip = ip
        self._cpu_context = cpu_context
        self._mnem = None
        self._root_mnem = None
        self.__insn = None

    def __deepcopy__(self, memo):
        # When we deep copy, clear out the __insn attribute so we don't
        # run into any serialization issues with Swig objects.
        deepcopy_method = self.__deepcopy__
        self.__deepcopy__ = None
        self.__insn = None
        copy = deepcopy(self, memo)
        self.__deepcopy__ = deepcopy_method
        return copy

    def __str__(self):
        return f"<{self.__class__.__name__} 0x{self.ip:08x} - {self.text}>"

    @property
    def _insn(self):
        if not self.__insn:
            insn = ida_ua.insn_t()
            ida_ua.decode_insn(insn, self.ip)
            if not insn:
                raise FunctionTracingError(f"Failed to decode instruction at 0x{self.ip:X}")
            self.__insn = insn
        return self.__insn

    @property
    def mnem(self) -> str:
        """Opcode mnemonic."""
        if not self._mnem:
            self._mnem = idc.print_insn_mnem(self.ip).lower()
        return self._mnem

    @property
    def root_mnem(self) -> str:
        """
        Opcode mnemonic without any extensions such as condition codes, data type, etc.

        e.g.
            MOVSEQ -> MOV
        """
        if self._root_mnem is None:
            self._root_mnem = self._insn.get_canon_mnem().lower()
        return self._root_mnem

    @property
    def text(self) -> str:
        """Disassembled code."""
        return idc.GetDisasm(self.ip)

    @property
    def is_terminal(self) -> bool:
        return ida_idp.is_ret_insn(self._insn)

    # TODO: Overwrite this in ARM to handle that special case.
    @property
    def operands(self) -> List[_operand_class]:
        # Calling insn_t() and decode_insn() is somewhat expensive and this function gets called a LOT
        # (with different instructions, so we are going to cache the operand indices globally.
        try:
            indices = self._operand_indices[self.ip]
        except KeyError:
            indices = []
            for idx, op in enumerate(self._insn.ops):
                if op.type == ida_ua.o_void:
                    break  # no more operands

                # IDA will sometimes create hidden or "fake" operands.
                # These are there to represent things like an implicit EAX register.
                # To help avoid confusion to the opcode developer, these fake operands will not be included.
                # TODO: Checking shown() may not work as expected.
                #   If things explode, go back to checking operand.is_hidden
                if op.shown():
                    indices.append((idx, op.type))

            self._operand_indices[self.ip] = indices

        return [self._operand_class(self._cpu_context, self.ip, idx, _type=type) for idx, type in indices]

    def _record_func_args(self):
        """
        Reset stack pointer reference and
        record function argument variables if we are executing the beginning of a function.
        """
        func_obj = ida_funcs.get_func(self.ip)
        if not func_obj or func_obj.start_ea != self.ip:
            return

        # Reset the sp_start
        self._cpu_context._sp_start = self._cpu_context.sp

        # Add the passed in arguments to the variables map.
        # (This also helps to standardize the argument names to "a*" instead of "arg_*")
        for arg in self._cpu_context.passed_in_args:
            addr = arg.addr
            # TODO: Support variables from registers?
            if addr is not None:
                if arg.is_stack:
                    try:
                        frame = ida_frame.get_frame(func_obj)
                        if not frame:
                            logger.warning(f"Failed to get frame for function argument: {repr(arg)}")
                            continue

                        # Getting member stack offset from name is more reliable then calculating
                        # it from the address.
                        member = ida_struct.get_member_by_name(frame, arg.name)
                        if not member:
                            logger.warning(f"Failed to get member for function argument: {repr(arg)}")
                            continue

                        self._cpu_context.variables.add(addr, frame_id=frame.id, stack_offset=member.soff)
                    except ValueError:
                        logger.warning(f"Failed to get stack information for function argument: {repr(arg)}")
                else:
                    self._cpu_context.variables.add(addr)

    def get_hooks(self, pre=True):
        """
        Retrieves callback hooks for the given instruction.
        :param pre: Whether to retrieve pre or post execution hooks.
        """
        return (
            self._cpu_context.emulator.get_instruction_hooks(self.ip, pre=pre)
            + self._cpu_context.emulator.get_instruction_hooks(self.mnem, pre=pre)
        )

    def _execute_hooks(self, pre=True):
        """
        Executes instructions hooks for the given start
        """
        for hook in self.get_hooks(pre=pre):
            try:
                hook(self._cpu_context, self)
            except RuntimeError:
                raise  # Allow RuntimeError exceptions to be thrown.
            except Exception as e:
                logger.debug("Failed to execute instruction hook with error: %s", e)

    def _execute(self):
        """
        Internal execute code used to execute the instruction itself and
        perform any custom tasks specific to the architecture.
        """
        emulator = self._cpu_context.emulator
        opcode_func = (
            emulator.get_opcode_hook(self.mnem)  # e.g. "bleq"
            or emulator.get_opcode_hook(self.root_mnem)  # e.g. "bl" from "bleq"
        )
        if opcode_func:
            opcode_func(self._cpu_context, self)
        else:
            logger.debug("%s instruction not implemented.", self.mnem)

    # TODO: All variable add_references() should be done here instead of in Operand class.
    def execute(self):
        """
        Emulate the instruction and store results in the context.
        """
        # Log a header line for debug messages of this instruction.
        # This is simpler and faster than trying to include the information at each log line
        logger.debug("[0x%X %03X] :: %s", self.ip, self._cpu_context.sp_diff, self.mnem)

        # Set instruction pointer to where we are currently executing.
        self._cpu_context.ip = self.ip

        # Extra processing if we are at the start of a function.
        self._record_func_args()

        # Run any pre-hooks first.
        self._execute_hooks(pre=True)

        # Execute the instruction.
        try:
            self._execute()
        except Exception:
            logger.exception("Failed to execute address 0x%X: %s", self.ip, self.text)

        # Record executed instruction.
        self._cpu_context.executed_instructions.append(self.ip)

        # Run any post-hooks.
        self._execute_hooks(pre=False)

        # Add a blank space to help visually separate logs for each instruction.
        logger.debug("  ")

        # After execution, set instruction pointer to next instruction assuming
        # standard code flow and if no jump was made.
        if self._cpu_context.ip == self.ip:
            self._cpu_context.ip = idc.next_head(self.ip)

    def execute_call_hooks(self, func_name, func_ea):
        """
        Collect call history and emulates the affects of the function call
        by executing call hooks.

        :param func_name: Name of the function (or empty string)
        :param func_ea: Address of function to call.
        """
        warnings.warn("execute_call_hooks() has been moved to ProcessorContext._execute_call()", DeprecationWarning)
        self._cpu_context._execute_call(func_name, func_ea, self.ip)



"""
Utility for interfacing with segments more efficiently.
"""

import numbers

import ida_bytes
import ida_segment
import idc

_cache = {}


def get_start(name_or_addr):
    """
    Retrieves the starting address for the segment containing the given name or address.

    :param string|int name_or_addr: either the name of a segment or an EA within a segment
    """
    if isinstance(name_or_addr, str):
        segment = ida_segment.get_segm_by_name(name_or_addr)
        if segment is None:
            raise AssertionError("could not find segment for {}".format(name_or_addr))
        return segment.start_ea
    elif isinstance(name_or_addr, numbers.Number):
        return idc.get_segm_attr(name_or_addr, idc.SEGATTR_START)
    else:
        raise ValueError("Invalid value: {}".format(name_or_addr))


def _obtain_bytes(start, end):
    """
    Obtain bytes efficiently, sets non-loaded bytes to \x00

    :param int start: starting address
    :param int end: ending address

    :return bytes: bytes contained within range
    """
    # Reconstruct the segment, account for bytes which are not loaded.
    # Can't use xrange() here because we can get a "Python int too large to conver to C long" error
    bytes_range = range(start, end)  # a range from start -> end
    return bytes(bytearray(ida_bytes.get_wide_byte(i) if idc.is_loaded(i) else 0 for i in bytes_range))


def get_bytes(name_or_addr):
    """
    Obtains segment bytes for the segment in which EA is contained or by segment name.
    This will be on demand and segment bytes will be cached if they have not already been obtained

    :param string|int name_or_addr: either the name of a segment or an EA within a segment

    :return bytes: bytes which are contained with the segment
    """
    seg_start = get_start(name_or_addr)
    seg_bytes = _cache.get(seg_start)
    if seg_bytes is None:
        seg_end = idc.get_segm_attr(seg_start, idc.SEGATTR_END)

        # This check will make an adjustment to seg_end based on whether the previous address has a value
        # or not.  In some instances, seg_end will throw us into an adjacent section which has data and
        # we'll end up getting bad values here.
        if idc.is_loaded(seg_end) and not idc.is_loaded(seg_end - 1):
            seg_end -= 1

        # Need to find the actual end address of the section data since IDA returns addresses that have
        # no data...
        while not idc.is_loaded(seg_end):
            seg_end -= 1

        seg_bytes = _obtain_bytes(seg_start, seg_end)
        _cache[seg_start] = seg_bytes

    return seg_bytes


def clear_cache():
    """
    Clears the internal cache of segment bytes.
    Calling this will be necessary if you have patched in new bytes into the IDB.
    """
    global _cache
    _cache = {}



"""
This module uses the idaapi.FlowChart object and extends it as well as idaapi.BasicBlock in order to add
functionality including Breadth-First and Depth-First chart traversal, locating a specific block within the
chart based on an EA, generating a list of all possible paths to a specified EA, etc.
"""

import functools
import logging
from operator import attrgetter
from typing import Optional, Set, Iterator

import ida_idaapi
import ida_gdl
import ida_funcs
import idautils

logger = logging.getLogger(__name__)


class PathNode(object):
    """
    Represents a linked-list of objects constituting a path from a specific node to the function entry point node.
    This object can also track cpu context up to a certain EA.
    """

    _cache = {}

    def __init__(self, bb: "BasicBlock", prev: Optional["PathNode"]):
        """
        Initialize a path node.

        :param bb: The underlying basic block for this node.
        :param prev: The parent node that points to this node.
        """
        self.bb = bb
        self.prev = prev

    @classmethod
    def from_cache(cls, bb: "BasicBlock", prev: Optional["PathNode"]):
        """Constructor that caches and reuses existing instances."""
        try:
            return cls._cache[(bb, prev)]
        except KeyError:
            path_node = cls(bb, prev)
            cls._cache[(bb, prev)] = path_node
            return path_node

    def __contains__(self, ea):
        return ea in self.bb

    def __repr__(self):
        return "PathNode({!r})".format(self.bb)

    def path(self):
        """Returns a list of PathNode objects represented by the linked list."""
        if self.prev:
            return self.prev.path() + [self]
        else:
            return [self]


@functools.total_ordering
class BasicBlock(object):
    """
    A reimplementation of  ida_gdl.BasicBlock object which has been extended with additional
    functionality.

    NOTE: This is a reimplementation of ida_gdl.BasicBlock. We are not inheriting from
    this class because it causes limitations for the IDA proxy feature.

    Additional functionality:
        - Ability to use BasicBlocks as hashable objects (ie: as dictionary keys)
        - Check if two BasicBlocks are equal (based on their start_ea)
        - Check if an EA is contained in a BasicBlock (ie: if ea in <BasicBlock>:)
        - Check length of block
        - Iterator addresses in block.
        - Iterator of paths leading to this block.
    """

    # This allows us to change the PathNode class in kordesii.utils.function_tracing.flowchart
    _PATHNODE_CLASS = PathNode

    def __init__(self, id, bb, fc):
        self._fc = fc
        self.id = id
        self.start_ea = bb.start_ea
        self.end_ea = bb.end_ea
        self.type = self._fc._q.calc_block_type(self.id)

    @classmethod
    def from_address(cls, ea):
        """Factory method for creating BasicBlock from an address."""
        fc = Flowchart.from_cache(ea)
        block = fc.find_block(ea)
        assert block, "Failed to get a BasicBlock"
        return block

    # region ida_gdl.BasicBlock functions

    def preds(self):
        """
        Iterates the predecessors list
        """
        q = self._fc._q
        for i in range(0, self._fc._q.npred(self.id)):
            yield self._fc[q.pred(self.id, i)]

    def succs(self):
        """
        Iterates the successors list
        """
        q = self._fc._q
        for i in range(0, q.nsucc(self.id)):
            yield self._fc[q.succ(self.id, i)]

    # endregion

    def __hash__(self):
        return self.start_ea

    def __repr__(self):
        return "<BasicBlock(start_ea=0x{:08X}, end_ea=0x{:08X})>".format(self.start_ea, self.end_ea)

    def __eq__(self, other):
        return self.start_ea == other.start_ea

    def __lt__(self, other):
        return self.start_ea < other.start_ea

    def __contains__(self, ea):
        return self.start_ea <= ea < self.end_ea

    def __len__(self):
        """Length of block is the number of instructions contained within."""
        return len(list(self.heads()))

    def heads(self, start=None, reverse=False):
        """
        Iterates all the heads within the given block.

        :param start: Start address (defaults to start_ea or end_ea)
        :param reverse: Direction to iterate

        :yields: Instruction addresses.

        :raises ValueError: If given start address it not in block.
        """
        if start and start not in self:
            raise ValueError("Start address 0x{:08X} is not in block: {!r}".format(start, self))

        if reverse:
            yield from reversed(list(idautils.Heads(self.start_ea, start or self.end_ea)))
        else:
            yield from idautils.Heads(start or self.start_ea, self.end_ea)

    def paths(self, _visited=None) -> Iterator[_PATHNODE_CLASS]:
        """
        Iterates the paths that lead to this block.

        :param _visited: Internally used.
        :yields: PathNode objects that represent the last entry of the path linked list.
        """
        if _visited is None:
            _visited = set()

        # Otherwise generate path nodes and cache results for next time.
        _visited.add(self.start_ea)

        parents = list(self.preds())
        if not parents:
            yield self._PATHNODE_CLASS.from_cache(self, prev=None)
        else:
            for parent in parents:
                if parent.start_ea in _visited:
                    continue

                # Create path nodes for each path of parent.
                for parent_path in parent.paths(_visited=_visited):
                    yield self._PATHNODE_CLASS.from_cache(self, prev=parent_path)

        _visited.remove(self.start_ea)

    def ancestors(self, _visited=None) -> Set["BasicBlock"]:
        """
        Returns a set of ancestor blocks for the given block.

        :param _visited: Internally used.

        :returns: Set of ancestor blocks.
        """
        if _visited is None:
            _visited = set()

        _visited.add(self)

        parents = set(parent for parent in self.preds() if parent not in _visited)
        ancestors = parents.union(*(parent.ancestors(_visited=_visited) for parent in parents))

        _visited.remove(self)

        return ancestors


class Flowchart(object):
    """
    Object containing the function graph generated by IDA.
    Implements the traversal of the function.

    NOTE: This is a reimplementation of ida_gdl.Flowchart. We are not inheriting from
    this class because it causes limitations for the IDA proxy feature.
    """

    _cache = {}

    # This allows us to change the BasicBlock class in kordesii.utils.function_tracing.flowchart
    _BASICBLOCK_CLASS = BasicBlock

    def __init__(self, func_ea):
        """
        Initializes Flowchart

        :param func_ea: An address in the function.
        """
        self.func_obj = ida_funcs.get_func(func_ea)
        self._q = ida_gdl.qflow_chart_t(
            "", self.func_obj, ida_idaapi.BADADDR, ida_idaapi.BADADDR, ida_gdl.FC_PREDS
        )

    @classmethod
    def from_cache(cls, func_ea):
        """Constructor that caches and reuses existing instances."""
        func = ida_funcs.get_func(func_ea)
        if not func:
            raise ValueError(f"{func_ea} is not within a function.")

        try:
            return cls._cache[func.start_ea]
        except KeyError:
            flowchart = cls(func.start_ea)
            cls._cache[func.start_ea] = flowchart
            return flowchart

    # region ida_gdl.Flowchart functions

    @property
    def size(self):
        return self._q.size()

    def refresh(self):
        self._q.refresh()

    def _getitem(self, index):
        return self._BASICBLOCK_CLASS(index, self._q[index], self)

    def __iter__(self):
        return (self._getitem(index) for index in range(0, self.size))

    def __getitem__(self, index):
        if index >= self.size:
            raise KeyError
        else:
            return self._getitem(index)

    # endregion

    def _traverse(self, start_ea=None, dfs=False):
        """
        Blind traversal of the graph.
        For each block, obtain the children (or blocks which are reachable
        from the current block), sort the children by their start_ea in ascending order, and "push" the list on to the
        front of the non_visisted blocks list.

        :param int start_ea: EA within a block from which to start traversing
        :param bool dfs: If true, traversal will be depth-first. If false, traversal will be breadth-first.

        :yield BasicBlock: function block object
        """
        # Set our flag to True if start_ea is none so we yield all blocks, else wait till we find the requested block
        block_found = start_ea is None
        non_visited = [self[0]]
        visited = set()
        while non_visited:
            cur_block = non_visited.pop(0)
            if hash(cur_block) in visited:
                continue

            visited.add(hash(cur_block))
            succs = sorted(cur_block.succs())
            if dfs:
                # [0:0] allows us to extend to the front
                non_visited[0:0] = succs
            else:
                non_visited.extend(succs)

            if not block_found:
                block_found = start_ea in cur_block

            if block_found:
                yield cur_block

    def _traverse_reverse(self, start_ea=None, dfs=False):
        """
        Perform a reverse traversal of the graph in depth-first/breadth-first manner where given a start node, traverse 1 complete
        path to the root node before following additional paths.

        :param int start_ea: EA within a block from which to start traversing
        :param bool dfs: If true, traversal will be depth-first. If false, traversal will be breadth-first.

        :yield: function block object
        """
        if start_ea:
            non_visited = [self.find_block(start_ea)]
        else:
            non_visited = list(sorted(self, key=attrgetter("start_ea")))[-1:]

        visited = set()
        while non_visited:
            cur_block = non_visited.pop(0)
            if hash(cur_block) in visited:
                continue

            visited.add(hash(cur_block))

            preds = sorted(cur_block.preds(), reverse=True)
            # For now, only consider predicates that are before the current block.
            # This helps to prevent cyclic loops.
            preds = [pred for pred in preds if pred < cur_block]
            if dfs:
                non_visited[0:0] = preds
            else:
                non_visited.extend(preds)

            yield cur_block

    def blocks(self, start=None, reverse=False, dfs=False):
        """
        Iterates over CustomBasicBlocks.

        :param int start: optional address to start iterating from.
        :param bool reverse: iterate in reverse
        :param bool dfs: If true, traversal will be depth-first. If false, traversal will be breadth-first.
        """
        if reverse:
            for cur_block in self._traverse_reverse(start, dfs=dfs):
                yield cur_block

        else:
            for cur_block in self._traverse(start, dfs=dfs):
                yield cur_block

    def heads(self, start=None, reverse=False, dfs=False):
        """
        Iterate over instructions in function blocks.

        :param int start: optional address to start iterating from.
        :param bool reverse: iterate in reverse
        :param bool dfs: If true, traversal of blocks will be depth-first.
            If false, traversal will be breadth-first.
        """
        _first_block = True
        for cur_block in self.blocks(start, reverse=reverse, dfs=dfs):
            if start and _first_block:
                yield from cur_block.heads(start, reverse=reverse)
            else:
                yield from cur_block.heads(reverse=reverse)

            _first_block = False

    def find_block(self, ea):
        """
        Locate a BasicBlock which contains the specified ea

        :param int ea: ea of interest

        :return: CustomBasicBlock object or None if not found.
        :rtype: BasicBlock
        """
        for block in self:
            if ea in block:
                return block

    def get_paths(self, ea) -> Iterator[PathNode]:
        """
        Given an EA, iterate over the paths to the EA.

        For usage example, see Emulator.iter_context_at()

        ..warning:: DO NOT WRAP THIS GENERATOR IN list()!!!  This generator will iterate all possible paths to the node containing
        the specified EA.  On functions containing large numbers of jumps, the number of paths grows exponentially and
        you WILL hit memory exhaustion limits, extremely slow run times, etc. Use extremely conservative constraints
        when iterating.  Nodes containing up to at least 32,768 paths are computed in a reasonably sane amount of time,
        though it probably doesn't make much sense to check this many paths for the data you are looking for.

        :param int ea: EA of interest

        :yield: a path to the object
        """
        # Obtain the block containing the EA of interest
        block = self.find_block(ea)

        # If block not found, then there are no paths to it.
        if not block:
            logger.debug("Unable to find block with ea: 0x{:08X}".format(ea))
            return

        yield from block.paths()



"""
IDA Pyro4 client used to run IDAPython remotely.
"""

import logging
import functools
import os
import subprocess
import sys
import tempfile
import time
import threading
import warnings

import dill
import Pyro4

import kordesii
from kordesii import ida_server
from kordesii import logutil


logger = logging.getLogger(__name__)

Pyro4.config.SERIALIZER = "dill"

_open_proxies = {}           # Maps package full name to Proxy object.
_port = None                 # Port currently used to access Proxies
_instance_running = False    # Whether a current IDA instance is running.
_run_in_ida_funcs = {}       # Registration of functions decorated as run_in_ida.


def run_in_ida(func):
    """Decorates a function to make available within the IDA proxy."""
    warnings.warn("IDA Proxy is deprecated in favor of using Dragodis.", DeprecationWarning)
    # Keep track of decorated functions, so we can allow decorated
    # functions to call each other.
    # But only allow for functions that are within the same module to
    # limit the amount that gets serialized.
    _run_in_ida_funcs.setdefault(func.__module__, {})
    _run_in_ida_funcs[func.__module__][func.__name__] = func

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if kordesii.in_ida:
            return func(*args, **kwargs)

        if not _instance_running:
            raise RuntimeError("IDA instance is not started!")

        # We get a pickling error if we allow Pyro4 to serialize the function.
        # But doing it ourselves seems to work fine.
        main_proxy = _open_proxies["main"]
        _globals = dill.dumps(_run_in_ida_funcs[func.__module__])
        _func = dill.dumps(func)
        ret, stdout, stderr = main_proxy.run_func(sys.path, _globals, _func, *args, **kwargs)
        sys.stdout.write(stdout)
        sys.stdout.flush()
        sys.stderr.write(stderr)
        sys.stderr.flush()
        if isinstance(ret, Exception):
            # Print out the stacktrace in IDA in a way that simulates a chained exception.
            for line in ret.ida_traceback:
                print(line, file=sys.stderr, end="")
            print(
                "\nThe above exception occurred in IDA and was the direct cause of the following exception:\n\n",
                file=sys.stderr, end=""
            )
            raise ret  # Exception occurred in IDA
        return ret

    return wrapper


def communicate(proc):
    """
    Drop the annoying swig memory leak warning. This is not a real concern for our purposes
    and only clutters the test results.
    """
    stdout, stderr = proc.communicate()
    ignore_str = "swig/python detected a memory leak of type 'std::out_of_range *', no destructor found."
    if stdout:
        for out_line in stdout.splitlines(True):
            if out_line.strip() == ignore_str:
                continue
            sys.stdout.write(out_line)

    if stderr:
        for out_line in stderr.splitlines(True):
            if out_line.strip() == ignore_str:
                continue
            sys.stderr.write(out_line)


class IDAModule(object):
    """
    Replicates an IDA module externally.

    Redirects all function and attributes to the proxied module within IDA.
    """

    def __init__(self, fullname):
        # Using self.__dict__ to avoid triggering the overwritten __setattr__()
        self.__dict__["fullname"] = fullname

    def __repr__(self):
        status = "open" if _instance_running else "closed"
        return "<kordesii.ida_client.IDAModule for {} : {}>".format(self.fullname, status)

    def __dir__(self):
        dir_ = []
        for name in dir(self._proxy):
            if name.startswith(("_pyro", "_Proxy")):
                continue
            if name.startswith("priv_"):
                name = name[4:]
            dir_.append(name)
        return dir_

    def __getattr__(self, item):
        if not _instance_running:
            raise AttributeError("IDA instance is closed.")

        # Hackery due to Pyro4 not allowing private methods and attributes.
        if item.startswith("_"):
            item = "priv" + item

        value = getattr(self._proxy, item)

        # Convert values that are ida_server.IDAClass or ida_server.IDAModule objects
        # to our ida_client counterpart.
        # NOTE: We can't just check if it's an instance of ida_server.IDAClass, because
        # the class was viewed as "__main__.IDAClass" on the IDA side.
        if isinstance(value, object):
            name = type(value).__name__
            if name == "IDAClass":
                value = IDAClass(value.object_id)
            elif name == "IDAModule":
                value = IDAModule(value.object_id)

        return value

    def __setattr__(self, key, value):
        if not _instance_running:
            raise AttributeError("IDA instance is closed.")

        # Hackery due to Pyro4 not allowing private methods and attributes.
        if key.startswith("_"):
            key = "priv" + key

        setattr(self._proxy, key, value)

    @property
    def _proxy(self):
        """
        Property to retrieve the Pyro4 proxy for the given module.
        Opens the proxy if not already.

        :return: proxy object
        :rtype: Pyro4.Proxy
        """
        if not _instance_running:
            raise AttributeError("IDA instance is closed.")
        try:
            return _open_proxies[self.fullname]
        except KeyError:
            try:
                proxy = Pyro4.Proxy("PYRO:{}@localhost:{}".format(self.fullname, _port))
                _open_proxies[self.fullname] = proxy

                # Replace any attributes that this module would have overwritten.
                self.__dict__["__doc__"] = getattr(proxy, "priv__doc__", self.__doc__)

                return proxy
            except Exception as e:
                raise RuntimeError("Cannot connect to IDA: {}".format(e))


class IDAClass(IDAModule):

    def __repr__(self):
        status = "open" if _instance_running else "closed"
        return "<kordesii.ida_client.IDAClass for {} : {}>".format(self.fullname, status)

    def __call__(self, *args, **kwargs):
        raise NotImplementedError(f"Initializing the class {self.fullname} is not supported.")


class IDALoader(object):
    """Import loader used to wrap any to-be proxied modules with IDAModule."""

    def load_module(self, fullname):
        try:
            return sys.modules[fullname]
        except KeyError:
            pass

        module = IDAModule(fullname)
        sys.modules[fullname] = module
        return module


class IDAFinder(object):
    """Import finder that passes IDALoader to any modules that should be proxied."""

    def find_module(self, fullname, path=None):
        # Fake the loader anytime we try to import a proxied module.
        if fullname in ida_server.proxied_modules:
            return IDALoader()
        return None


class IDA(object):
    """Allows running IDAPython script outside if IDA!"""

    _internal_extensions = [".id0", ".id1", ".id2", ".nam", ".til"]

    def __init__(self, input_path, keep_idb=False):
        """
        :param input_path: Path to file to examine.
        :param keep_idb: Whether to keep the idb around after stopping.
        """
        warnings.warn("IDA Proxy is deprecated in favor of using Dragodis.", DeprecationWarning)
        self.input_path = input_path
        # Keep idb if requested or an idb already was there.
        self.keep_idb = keep_idb or os.path.exists(input_path + ".idb") or os.path.exists(input_path + ".i64")
        self._orig_path = None
        self._orig_modules = None
        self._orig_meta_path = None
        self._thread = None
        self._process = None
        self._started = False

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    def __del__(self):
        self.stop()

    @staticmethod
    def _make_port_file():
        """
        Creates a temporary file to transfer the connection port.

        The path to this file is transferred over the command line
        and the Pyro4 server (run in IDA) writes the connection port
        to it when it's been generated.
        """
        temp = tempfile.NamedTemporaryFile("rb", delete=False)
        temp.file.close()
        return temp.name

    @staticmethod
    def _get_port(port_path):
        """
        Get the saved connection port. Occasionally it takes
        a couple seconds for IDA to start up and write the port.

        :raises: An exception if occurred in ida_server.py
        """
        attempts = 0
        while attempts < 100:
            with open(port_path, "rb") as f:
                data = f.read()

            if data:
                ret = dill.loads(data)
                if isinstance(ret, Exception):
                    raise ret  # This exception occurred within IDA during setup.
                return ret

            attempts += 1
            time.sleep(0.25)

        raise RuntimeError("Unable to get IDA server port file.")

    def _open_in_ida(self):
        """
        Checks if input file is already open in idea by checking for the existance of
        internal runtime files.
        """
        return any(os.path.exists(self.input_path + ext) for ext in self._internal_extensions)

    def start(self):
        """
        Starts IDA proxy server and sets up import hooks.

        :raises RuntimeError: If we fail to start or connect to IDA.
        """
        global _instance_running
        global _open_proxies
        global _port

        if _instance_running:
            raise ValueError("IDA instance already started!")

        if not os.path.exists(self.input_path):
            raise ValueError("Unable to find: {}".format(self.input_path))

        if self._open_in_ida():
            raise ValueError(
                "Another process has {} open in IDA. "
                "Please close the other process and/or cleanup the internal runtime files.".format(self.input_path)
            )

        # Start log listener (if not already started)
        if not logutil.listen_port:
            kordesii.setup_logging()
            assert logutil.listen_port

        # Start up IDA server.
        ida_exe = kordesii.find_ida(kordesii.is_64_bit(self.input_path))
        script_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "ida_server.py")
        port_path = self._make_port_file()
        command = [
            ida_exe,
            "-A",
            '-S"\"{}\" {} {} {} exit"'.format(
                script_path, logging.root.getEffectiveLevel(), logutil.listen_port, port_path),
            # '-L"{}.log"'.format(self.input_path),   # Uncomment if debugging ida_server.py
            '"{}"'.format(self.input_path),
        ]
        command = " ".join(command)
        logger.debug("Running command: {}".format(command))

        env = dict(os.environ)
        env["PYTHONPATH"] = ""
        process = subprocess.Popen(command, env=env, shell=sys.platform != "win32")

        thread = threading.Thread(target=communicate, args=(process,))
        thread.daemon = True
        thread.start()

        try:
            # Wait for port to arrive then retrieve it.
            logger.debug("Retrieving port...")
            _port = self._get_port(port_path)
            if os.path.exists(port_path):
                os.unlink(port_path)
            logger.debug("IDA Server setup on port {}".format(_port))

            # Pull main proxy.
            main_proxy = Pyro4.Proxy("PYRO:main@localhost:{}".format(_port))
            assert main_proxy.ping()  # Safety check
        except Exception:
            process.kill()
            thread.join()
            self._cleanup_files()
            raise

        _open_proxies["main"] = main_proxy
        self._process = process
        self._thread = thread
        self._orig_path = list(sys.path)
        self._orig_modules = dict(sys.modules)
        self._orig_meta_path = list(sys.meta_path)

        # Hook import paths to use proxy for internal ida module.
        sys.path.append(os.path.join(os.path.dirname(ida_exe), "python"))
        sys.modules["__main__"].IDAPYTHON_COMPAT_695_API = False
        sys.meta_path.insert(0, IDAFinder())

        _instance_running = True
        self._started = True

    def _stop_server(self):
        """Stop the IDA server."""
        # The first open proxy is our main proxy containing the stop_daemon() function we need to call.
        main_proxy = _open_proxies["main"]
        main_proxy.stop_daemon()

    def _cleanup_files(self):
        """Cleans up any remaining IDA files."""
        extensions = self._internal_extensions[:]
        if not self.keep_idb:
            extensions += [".idb", ".i64"]
        for extension in extensions:
            file_path = self.input_path + extension
            if os.path.exists(file_path):
                try:
                    os.unlink(file_path)
                except Exception as e:
                    logger.warning("Failed to remove {} with error: {}".format(file_path, e))

    def stop(self):
        """Stops IDA proxy server and removes import hooks."""
        global _open_proxies
        global _instance_running

        if not self._started:
            return

        try:
            # Restore import paths.
            sys.path = self._orig_path
            sys.modules = self._orig_modules
            sys.meta_path = self._orig_meta_path

            # Stop the proxies
            self._stop_server()
            for proxy in _open_proxies.values():
                proxy._pyroRelease()
                # Bug in Pyro4 teardown causes a warning on gc if
                # this attribute still exists.
                delattr(proxy, "_pyroConnection")
            _open_proxies = {}

            # Stop thread
            self._thread.join(5)
            if self._thread.is_alive():
                self._process.kill()
                self._thread.join(5)

            _instance_running = False
            self._started = False
        finally:
            # Ensure any remaining IDA files are removed.
            self._cleanup_files()



"""
CPU Emulator Built-in Windows API functions

These functions are used to emulate the effects of "useful" Windows API functions.

Add any builtin functions that need to be handled below.  The function should be declared as such

# Using the same function for multiple API functions
@builtin_func("GetSystemDirectoryA")
@builtin_func("GetSystemDirectoryW")
#typespec("UINT GetSystemDirectoryA(char* lpBuffer, UINT uSize);")
def memmove(cpu_context, func_name, func_args):
    logger.debug("Emulating GetSystemDirectoryA")

"""
import logging
import random

from . import win_constants as wc
from ... import actions, utils, constants, objects
from ...call_hooks import builtin_func

logger = logging.getLogger(__name__)


def _get_reg_key(cpu_context, root_key_handle, sub_key_ptr, wide=False):
    """
    Retrieves the root key and sub key strings from the given hKey and lpSubKey arguments.

    :param root_key_handle: The hKey argument (the HKEY handle for the root key)
    :param sub_key_ptr: The lpSubKey argument (the pointer to the sub key string)
    :param wide: Whether strings are utf16 or utf8
    """
    # Make sure root_key_handle is 32-bit
    root_key_handle &= 0xFFFFFFFF
    # Get root key
    try:
        reg_key = cpu_context.objects[root_key_handle]
        root_key = reg_key.path
    except KeyError:
        if root_key_handle in wc.RegistryKey.__members__.values():
            # root key is a predefined enum.
            root_key = wc.RegistryKey(root_key_handle).name
        else:
            logger.warning("Invalid registry key 0x%X, using hex string.", root_key_handle)
            root_key = hex(root_key_handle)

    # Get sub key
    if not sub_key_ptr:
        sub_key = None
    else:
        sub_key = cpu_context.read_data(
            sub_key_ptr, data_type=constants.WIDE_STRING if wide else constants.STRING
        ).decode("utf-16-le" if wide else "utf8")

    return root_key, sub_key


@builtin_func("RegOpenKeyA")
@builtin_func("RegOpenKeyW")
@builtin_func("RegOpenKeyExA")
@builtin_func("RegOpenKeyExW")
@builtin_func("RegCreateKeyA")
@builtin_func("RegCreateKeyW")
@builtin_func("RegCreateKeyExA")
@builtin_func("RegCreateKeyExW")
#typespec("LSTATUS RegOpenKeyA(HKEY hKey, LPCSTR lpSubKey, PHKEY phkResult)")
#typespec("int RegCreateKeyA(HKEY hKey, char* lpSubKey, HKEY* phkResult);")
#typespec("int RegCreateKeyW(HKEY hKey, wchar* lpSubKey, HKEY* phkResult);")
#typespec("int RegCreateKeyExA(HKEY hKey, char* lpSubKey, DWORD, char* lpClass, DWORD dwOptions, REGSAM samDesired, const SECURITY_ATTRIBUTES* lpSecurityAttributes, HKEY* phkResult, DWORD* lpdwDisposition);")
#typespec("int RegCreateKeyExA(HKEY hKey, wchar* lpSubKey, DWORD, wchar* lpClass, DWORD dwOptions, REGSAM samDesired, const SECURITY_ATTRIBUTES* lpSecurityAttributes, HKEY* phkResult, DWORD* lpdwDisposition);")
def reg_open_key(cpu_context, func_name, func_args):
    """
    Opens/Creates the specified registry key

    We are merging RegOpenKey* and RegCreateKey* because their effects are essentially same except
    for the extra disposition argument.
    """
    wide = func_name.endswith("W")
    root_key_handle, sub_key_ptr, *rest = func_args
    if func_name.startswith("RegCreateKeyEx"):
        result_ptr = rest[-2]  # RegCreateKeyEx has an extra disposition argument at the end.
    else:
        result_ptr = rest[-1]  # result pointer is last argument in both Ex and non-Ex versions.

    root_key, sub_key = _get_reg_key(cpu_context, root_key_handle, sub_key_ptr, wide)

    # If subkey is null or an empty string, result is the same as the root key handle.
    if not sub_key:
        disposition = wc.REG_OPENED_EXISTING_KEY
        cpu_context.mem_write(
            result_ptr, utils.struct_pack(root_key_handle, width=cpu_context.byteness)
        )
        logger.debug("Opening existing registry key: %s", root_key)

    # Otherwise create new RegKey object.
    else:
        disposition = wc.REG_CREATED_NEW_KEY
        path = "\\".join([root_key, sub_key])
        handle = cpu_context.objects.alloc()
        cpu_context.mem_write(
            result_ptr, utils.struct_pack(handle, width=cpu_context.byteness)
        )
        cpu_context.actions.add(
            actions.RegKeyOpened(cpu_context.ip, handle, path, root_key, sub_key)
        )
        logger.debug("Opening registry key: %s", path)

    # Need to report disposition if RegCreateKeyEx*
    if func_name.startswith("RegCreateKeyEx"):
        disposition_ptr = func_args[-1]
        if disposition_ptr:
            cpu_context.write_data(disposition_ptr, disposition, data_type=constants.DWORD)

    return wc.ERROR_SUCCESS


@builtin_func("RegDeleteKeyA")
@builtin_func("RegDeleteKeyW")
@builtin_func("RegDeleteKeyExA")
@builtin_func("RegDeleteKeyExW")
#typespec("int RegDeleteKeyA(HKEY hKey, char* lpSubKey);")
#typespec("int RegDeleteKeyW(HKEY hKey, wchar* lpSubKey);")
#typespec("int RegDeleteKeyExA(HKEY hKey, char* lpSubKey, REGSAM samDesired, DWORD Reserved);")
#typespec("int RegDeleteKeyExW(HKEY hKey, wchar* lpSubKey, REGSAM samDesired, DWORD Reserved);")
def reg_delete_key(cpu_context, func_name, func_args):
    """
    Deletes a subkey and its values.
    """
    wide = func_name.endswith("W")
    root_key_handle, sub_key_ptr, *_ = func_args

    root_key, sub_key = _get_reg_key(cpu_context, root_key_handle, sub_key_ptr, wide)

    if not sub_key:
        logger.warning("Sub key is null")
        return

    path = "\\".join([root_key, sub_key])
    cpu_context.actions.add(actions.RegKeyDeleted(cpu_context.ip, root_key_handle, path))
    logger.debug("Deleting registry key: %s", path)

    return wc.ERROR_SUCCESS


@builtin_func("RegDeleteKeyValueA")
@builtin_func("RegDeleteKeyValueW")
#typespec("int RegDeleteKeyValueA(HKEY hKey, char* lpSubKey, char* lpValueName);")
#typespec("int RegDeleteKeyValueW(HKEY hKey, wchar* lpSubKey, wchar* lpValueName);")
def reg_delete_key_value(cpu_context, func_name, func_args):
    """
    Removes the specified value from the specified registry key and subkey.
    """
    wide = func_name.endswith("W")
    root_key_handle, sub_key_ptr, value_name_ptr = func_args

    root_key, sub_key = _get_reg_key(cpu_context, root_key_handle, sub_key_ptr, wide)
    value_name = cpu_context.read_data(
        value_name_ptr, data_type=constants.WIDE_STRING if wide else constants.STRING
    ).decode("utf-16-le" if wide else "utf8")

    path = "\\".join([root_key, sub_key or ""])
    cpu_context.actions.add(
        actions.RegKeyValueDeleted(cpu_context.ip, root_key_handle, path, value_name)
    )
    logger.debug("Deleting value %s from registry key %s", value_name, path)

    return wc.ERROR_SUCCESS


@builtin_func("RegDeleteValueA")
@builtin_func("RegDeleteValueW")
#typespec("int RegDeleteValueA(HKEY hKey, char* lpValueName);")
#typespec("int RegDeleteValueW(HKEY hKey, wchar* lpValueName);")
def reg_delete_value(cpu_context, func_name, func_args):
    """
    Removes a named value from the specified registry key.
    """
    wide = func_name.endswith("W")
    root_key_handle, value_name_ptr = func_args

    root_key, _ = _get_reg_key(cpu_context, root_key_handle, 0, wide)
    value_name = cpu_context.read_data(
        value_name_ptr, data_type=constants.WIDE_STRING if wide else constants.STRING
    ).decode("utf-16-le" if wide else "utf8")

    path = root_key
    cpu_context.actions.add(
        actions.RegKeyValueDeleted(cpu_context.ip, root_key_handle, path, value_name)
    )
    logger.debug("Deleting value %s from registry key %s", value_name, path)

    return wc.ERROR_SUCCESS


@builtin_func("RegSetValueA")
@builtin_func("RegSetValueW")
@builtin_func("RegSetValueExA")
@builtin_func("RegSetValueExW")
#typespec("int RegSetValueA(HKEY hKey, char* lpSubKey, DWORD dwType, char* lpData, DWORD cbData);")
#typespec("int RegSetValueA(HKEY hKey, wchar* lpSubKey, DWORD dwType, wchar* lpData, DWORD cbData);")
#typespec("int RegSetValueExA(HKEY, char* lpValueName, DWORD Reserved, DWORD dwType, const BYTE *lpData, DWORD cbData);")
#typespec("int RegSetValueExA(HKEY, wchar* lpValueName, DWORD Reserved, DWORD dwType, const BYTE *lpData, DWORD cbData);")
def reg_set_value(cpu_context, func_name, func_args):
    """
    Sets the data for the default or unnamed value of a specified registry key.
    """
    wide = func_name.endswith("W")
    root_key_handle, sub_key_ptr, *_, data_type, data_ptr, data_size = func_args

    root_key, sub_key = _get_reg_key(cpu_context, root_key_handle, sub_key_ptr, wide)

    path = "\\".join([root_key, sub_key or ""])
    data_type = wc.RegistryDataType(data_type)

    # Data is a null-terminated string
    if data_type.name in ("REG_SZ", "REG_LINK", "REG_EXPAND_SZ"):
        data = cpu_context.read_data(data_ptr, data_type=constants.STRING)
        data = data.decode("utf8")

    # Data is a sequence of null-terminated strings, terminated by an empty string (\0)
    elif data_type.name == "REG_MULTI_SZ":
        # Since the last string is empty, this terminates as a double null,
        # which allows us to use WIDE_STRING.
        data = cpu_context.read_data(data_ptr, data_type=constants.WIDE_STRING)
        data = [string.decode("utf8") for string in data.split(b'\x00') if string]

    elif data_type.name == "REG_BINARY":
        data = cpu_context.read_data(data_ptr, size=data_size)

    elif data_type.name == "REG_DWORD":
        data = cpu_context.read_data(data_ptr, data_type=constants.DWORD)

    elif data_type.name == "REG_QWORD":
        data = cpu_context.read_data(data_ptr, data_type=constants.QWORD)

    elif data_type.name == "REG_NONE":
        data = None

    else:
        raise NotImplementedError(f"Unsupported data type: {data_type.name}")

    cpu_context.actions.add(
        actions.RegKeyValueSet(cpu_context.ip, root_key_handle, path, data_type.name, data)
    )
    logger.debug("Setting value %r to registry key %s", data, path)

    return wc.ERROR_SUCCESS


@builtin_func("OpenSCManagerA")
@builtin_func("OpenSCManagerW")
#typespec("HANDLE OpenSCManagerA(char* lpMachineName, char* lpDatabasename, DWORD dwDesiredAccess);")
#typespec("HANDLE OpenSCManagerA(wchar* lpMachineName, wchar* lpDatabasename, DWORD dwDesiredAccess);")
def open_sc_manager(cpu_context, func_name, func_args):
    """
    Establishes a connection to the service control manager on the specified computer
    and opens the specified service control manager database.
    """
    return random.randint(wc.MIN_HANDLE, wc.MAX_HANDLE)


@builtin_func("CreateServiceA")
@builtin_func("CreateServiceW")
#typespec("HANDLE CreateServiceA(HANDLE hSCManager, char* lpServiceName, char* lpDisplayName, DWORD dwDesiredAccess, DWORD dwServiceType, DWORD dwStartType, DWORD dwErrorControl, char* lpBinaryPathName, char* lpLoadOrderGroup, DWORD* lpdwTagId, char* lpDependencies, char* lpServiceStartName, char* lpPassword);")
#typespec("HANDLE CreateServiceW(HANDLE hSCManager, wchar* lpServiceName, wchar* lpDisplayName, DWORD dwDesiredAccess, DWORD dwServiceType, DWORD dwStartType, DWORD dwErrorControl, wchar* lpBinaryPathName, wchar* lpLoadOrderGroup, DWORD* lpdwTagId, wchar* lpDependencies, wchar* lpServiceStartName, wchar* lpPassword);")
def create_service(cpu_context, func_name, func_args):
    """
    Creates a service object and adds it to the specified service control manager database.
    """
    wide = func_name.endswith("W")
    service_name_ptr = func_args[1]
    display_name_ptr = func_args[2]
    desired_access = func_args[3]
    service_type = func_args[4]
    start_type = func_args[5]
    binary_path_ptr = func_args[7]

    service_name = cpu_context.read_data(service_name_ptr).decode("utf-16-le" if wide else "utf8")
    display_name = cpu_context.read_data(display_name_ptr).decode("utf-16-le" if wide else "utf8")
    binary_path = cpu_context.read_data(binary_path_ptr).decode("utf-16-le" if wide else "utf8")
    desired_access = wc.ServiceAccess(desired_access)
    service_type = wc.ServiceType(service_type)
    start_type = wc.ServiceStart(start_type)

    # TODO: Determine if a new handle is always created.
    handle = cpu_context.objects.get_or_alloc(objects.Service, name=service_name)

    action = actions.ServiceCreated(
        ip=cpu_context.ip,
        handle=handle,
        name=service_name,
        access=desired_access,
        service_type=service_type,
        start_type=start_type,
        display_name=display_name,
        binary_path=binary_path
    )
    cpu_context.actions.add(action)
    logger.debug("Created: %r", action)
    return handle


@builtin_func("OpenServiceA")
@builtin_func("OpenServiceW")
#typespec("SC_HANDLE OpenServiceA(SC_HANDLE hSCManager, LPCSTR lpServiceName, DWORD dwDesiredAccess);")
#typespec("SC_HANDLE OpenServiceW(SC_HANDLE hSCManager, LPCWSTR lpServiceName, DWORD dwDesiredAccess);")
def open_service(cpu_context, func_name, func_args):
    """
    Open a service
    """
    wide = func_name.endswith("W")
    service_name_ptr = func_args[1]
    service_name = cpu_context.read_data(service_name_ptr).decode("utf-16-le" if wide else "utf8")

    # TODO: Determine if a new handle is always created.
    handle = cpu_context.objects.get_or_alloc(objects.Service, name=service_name)

    action = actions.ServiceOpened(
        ip=cpu_context.ip,
        handle=handle,
        name=service_name
    )
    cpu_context.actions.add(action)
    logger.debug("Opened: %r", action)
    return handle


@builtin_func("DeleteService")
def delete_service(cpu_context, func_name, func_args):
    """
    BOOL DeleteService(SC_HANDLE hService);

    Delete service by handle
    """
    handle = func_args[0]

    if not handle:
        logger.warning(f"Service with handle {hex(handle)} does not exist.")
        
    try:
        action = actions.ServiceDeleted(
            ip=cpu_context.ip,
            handle=handle
        )
        cpu_context.actions.add(action)
        logger.debug("Deleted: %r", action)
    except ValueError as err:
        logger.debug("DeleteService: %s", err)

    return 1


@builtin_func("ChangeServiceConfig2A")
@builtin_func("ChangeServiceConfig2W")
#typespec("BOOL ChangeServiceConfig2A(SC_HANDLE hService, DWORD dwInfoLevel, LPVOID lpInfo);")
#typespec("BOOL ChangeServiceConfig2W(SC_HANDLE hService, DWORD dwInfoLevel, LPVOID lpInfo);")
def change_service_config(cpu_context, func_name, func_args):
    """
    Changes service configuration.

    NOTE: For now, only concerned with info level 1 for service description
    """
    wide = func_name.endswith("W")
    ptr_type = constants.QWORD if utils.get_bits() == 64 else constants.DWORD
    handle = func_args[0]
    info_level = func_args[1]
    info_ptr = func_args[2]
    ptr_value = cpu_context.read_data(info_ptr, data_type=ptr_type)

    if not handle:
        logger.warning(f"Service with handle {hex(handle)} does not exist.")

    if info_level == 1: # SERVICE_CONFIG_DESCRIPTION
        service_description = cpu_context.read_data(ptr_value).decode("utf-16-le" if wide else "utf8")

        action = actions.ServiceDescriptionChanged(
            ip=cpu_context.ip,
            handle=handle,
            description=service_description
        )
        cpu_context.actions.add(action)
        logger.debug(f"Description for service with handle {hex(handle)} changed: {service_description}")

    return 1



"""
This module extends the flowchart module in kordesii.utils in order to add ProcessorContext
tracking and emulation.
"""
from __future__ import annotations
import logging
import warnings
from copy import deepcopy
from typing import TYPE_CHECKING, Optional

import idautils

from kordesii.utils import flowchart

if TYPE_CHECKING:
    from .cpu_context import ProcessorContext


logger = logging.getLogger(__name__)


# TODO: Refactor this so we no longer need to have a separate version of PathNode.
class PathNode(flowchart.PathNode):
    """
    Extends original PathNode to add ability to get cpu_context.
    """

    _cache = {}

    def __init__(self, bb: BasicBlock, prev: Optional[PathNode]):
        """
        Initialize a path node.

        :param bb: The underlying basic block for this node.
        :param prev: The parent node that points to this node.
        """
        super().__init__(bb, prev)
        # TODO: Add caching for multiple init_contexts?
        self._context = None
        self._context_ea = None  # ea that the context has been filled to (but not including)
        self._init_context = None  # the context used at the starting path
        self._call_depth = 0       # the number of calls deep we are allowed to emulate

    def cpu_context(self, ea=None, *, call_depth: int = 0, init_context: ProcessorContext) -> ProcessorContext:
        """
        Returns the cpu context filled to (but not including) the specified ea.

        :param int ea: address of interest (defaults to the last ea of the block)
        :param call_depth: Number of function calls we are allowed to emulate into.
            When we hit our limit (depth is 0), emulation will no longer jump into function calls.
            (Defaults to not emulating into any function calls.)
            NOTE: This does not affect call hooks.
        :param init_context: Initial context to use for the start of the path. (required)

        :return cpu_context.ProcessorContext: cpu context
        """
        if ea is not None and ea not in self.bb:
            raise KeyError(
                "Provided address 0x{:X} not in this block "
                "(0x{:X} :: 0x{:X})".format(ea, self.bb.start_ea, self.bb.end_ea)
            )

        # Determine address to stop computing.
        if ea is None:
            end = self.bb.end_ea  # end_ea of a BasicBlock is the first address after the last instruction.
        else:
            end = ea

        # Determine if we need to force the creation of a new context if we have a different init_context
        # or call_depth.
        new_init_context = self._init_context != init_context or self._call_depth != call_depth
        self._init_context = init_context
        self._call_depth = call_depth

        assert end is not None
        # Fill context up to requested endpoint.
        if self._context_ea != end or new_init_context:
            # Create context if:
            #   - not created
            #   - current context goes past requested ea
            #   - given init_context/call_depth is different from the previously given init_context/call_depth.
            if not self._context or self._context_ea > end or new_init_context:
                # Need to check if there is a prev, if not, then we need to create a default context here...
                if self.prev:
                    self._context = self.prev.cpu_context(call_depth=call_depth, init_context=init_context)
                    # Modify the context for the current branch if required
                    self._context.prep_for_branch(self.bb.start_ea)
                else:
                    self._context = deepcopy(init_context)

                self._context_ea = self.bb.start_ea

            if self._context_ea != end:
                # Fill context up to requested ea.
                logger.debug("Emulating instructions 0x%08X -> 0x%08X", self._context_ea, end)
                for ip in idautils.Heads(self._context_ea, end):
                    self._context.execute(ip, call_depth=call_depth)

            self._context_ea = end

        # Set the next instruction pointer to be the end instruction that we did NOT execute.
        self._context.ip = end

        return deepcopy(self._context)


class BasicBlock(flowchart.BasicBlock):
    """
    Inherited version of BasicBlock that uses our modified PathNode for paths.
    """
    _PATHNODE_CLASS = PathNode


class Flowchart(flowchart.Flowchart):
    """
    Inherited version of Flowchart that uses our modified BasicBlock.
    """
    _cache = {}
    _BASICBLOCK_CLASS = BasicBlock


class FlowChart(Flowchart):
    def __init__(self, *args, **kwargs):
        warnings.warn("FlowChart has been renamed to Flowchart", DeprecationWarning)
        super(FlowChart, self).__init__(*args, **kwargs)




"""
Function tracing utility which will trace a function to a given EA and provide the value stored in a given operand.idaapi.__struct_unpack_table

This will essentially emulate, on a very primitive level, a x86(_64) cpu during execution and will maintain the stack,
registers, etc.
"""

import logging
import warnings
from builtins import object

from .emulator import Emulator


warnings.warn(
    "The function_tracer module is deprecated. Please create and use an instance of Emulator instead.", DeprecationWarning)

logger = logging.getLogger(__name__)


# Create a global Emulator to keep backwards compatibility with the legacy FunctionTracer and TracerCache.
try:
    _emulator = Emulator()
except NotImplementedError:
    _emulator = None


def FunctionTracer(func_ea):
    warnings.warn("FunctionTracer class is deprecated. Please create an instance of Emulator instead.", DeprecationWarning)
    return _emulator


class TracerCache(object):
    def __init__(self):
        warnings.warn("TracerCache is deprecated, please use an instance of Emulator instead.", DeprecationWarning)

    def get(self, ea, default="NOTSET"):
        return _emulator

    def hook(self, name_or_start_ea, func):
        _emulator.hook_call(name_or_start_ea, func)

    def clear_hooks(self):
        _emulator.reset_hooks()


def get_tracer(ea, default="NOTSET"):
    warnings.warn("get_tracer() is deprecated. Please create an instance of Emulator instead.", DeprecationWarning)
    return _emulator


def hook_tracers(name_or_start_ea, func):
    warnings.warn(
        "hook_tracers() is deprecated. Please call hook_call() on an instance of Emulator instead.", DeprecationWarning)
    _emulator.hook_call(name_or_start_ea, func)


def clear_hooks():
    warnings.warn(
        "clear_hooks() is deprecated. Please call clear_hooks() on an instance of Emulator instead.", DeprecationWarning)
    _emulator.reset_hooks()



"""Stores contants"""


# Define some datatypes that can be used for input to get the operand value
STRING = 0  # Represents a C-like string which is null terminated
WIDE_STRING = 1  # Represents a wide C-like string which is null terminated
BYTE_STRING = 2  # List a sequent of bytes of a specified size
BYTE = 3  # byte value
WORD = 4  # WORD value
DWORD = 5  # DWORD value
QWORD = 6  # QWORD value

STRING_TYPES = (STRING, WIDE_STRING, BYTE_STRING)
INT_TYPES = (BYTE, WORD, DWORD, QWORD)
DATA_TYPES = STRING_TYPES + INT_TYPES

SINGLE = 1  # Represents single precision for converting float to int or int to float
DOUBLE = 2  # Represents double precision for converting float to int or int to float



"""
This is a partial port of IDA's module/arm/arm.hpp from the Hex-Rays SDK.
(Hex-Rays does not officially expose this module in IDAPython.)

NOTE: NEON is currently being ignored. We may add it in the future.
"""
from enum import IntEnum, auto

import ida_allins
import ida_ua


# ARM insn.auxpref bits
aux_cond =        0x0001  # set condition codes (S postfix is required)
aux_byte =        0x0002  # byte transfer (B postfix is required)
aux_npriv =       0x0004  # non-privileged transfer (T postfix is required)
aux_regsh =       0x0008  # shift count is held in a register (see o_shreg)
aux_negoff =      0x0010  # memory offset is negated in LDR,STR
aux_immcarry =    0x0010  # carry flag is set to bit 31 of the immediate operand (see may_set_carry)
aux_wback =       0x0020  # write back (! postfix is required)
aux_wbackldm =    0x0040  # write back for LDM/STM (! postfix is required)
aux_postidx =     0x0080  # post-indexed mode in LDR,STR
aux_ltrans =      0x0100  # long transfer in LDC/STC (L postfix is required)
aux_wimm =        0x0200  # thumb32 wide encoding of immediate constant (MOVW)
aux_sb =          0x0400  # signed byte (SB postfix)
aux_sh =          0x0800  # signed halfword (SH postfix)
aux_sw =          (aux_sb|aux_sh)  # signed word (SW postfix)
aux_h =           0x1000  # halfword (H postfix)
aux_x =           (aux_h|aux_byte)  # doubleword (X postfix in A64)
aux_d =           aux_x   # dual (D postfix in A32/T32)
aux_p =           0x2000  # priviledged (P postfix)
aux_coproc =      0x4000  # coprocessor instruction
aux_wide =        0x8000  # wide (32-bit) thumb instruction (.W suffix)
aux_pac =        0x10000  # Pointer Authentication Code instruction (see PAC_ flags)
aux_ns =         0x20000  # non-secure branch (NS suffix)

# assembler flags
UAS_GNU =         0x0001  # GNU assembler
UAS_LEGACY =      0x0002  # Legacy (pre-UAL) assembler


# Operand Types:

o_shreg = ida_ua.o_idpspec0   # Shifted register

#define shtype          specflag2          //  op.shtype - shift type
#define shreg(x)        uchar(x.specflag1) //  op.shreg  - shift register
#define shcnt           value              //  op.shcnt  - shift counter

#define ishtype         specflag2          // o_imm - shift type
#define ishcnt          specval            // o_imm - shift counter

def shreg(op: ida_ua.op_t) -> int:
    """op.shreg - shift register"""
    return op.specflag1 & 0xFF

secreg = shreg  # o_phrase - the second register is here

#define ralign          specflag3          # o_phrase, o_displ: NEON alignment (power-of-two bytes, i.e. 8*(1<<a))
                                           # minimal alignment is 16 (a==1)

#define simd_sz         specflag1          # o_reg: SIMD vector element size
                                           # 0=scalar, 1=8 bits, 2=16 bits, 3=32 bits, 4=64 bits, 5=128 bits)
                                           # number of lanes is derived from the vector size (dtype)
#define simd_idx        specflag3          // o_reg: SIMD scalar index plus 1 (Vn.H[i])

# o_phrase: the second register is held in secreg (specflag1)
#           the shift type is in shtype (specflag2)
#           the shift counter is in shcnt (value)


o_reglist = ida_ua.o_idpspec1    # Register list (for LDM/STM)
#define reglist         specval            // The list is in op.specval
#define uforce          specflag1          // PSR & force user bit (^ suffix)

o_creglist = ida_ua.o_idpspec2   # Coprocessor register list (for CDP)
#define CRd             reg                //
#define CRn             specflag1          //
#define CRm             specflag2          //

o_creg = ida_ua.o_idpspec3       # Coprocessor register (for LDC/STC)

o_fpreglist = ida_ua.o_idpspec4  # Floating point register list
#define fpregstart      reg                // First register
#define fpregcnt        value              // number of registers; 0: single register (NEON scalar)
#define fpregstep       specflag2          // register spacing (0: {Dd, Dd+1,... }, 1: {Dd, Dd+2, ...} etc)
#define fpregindex      specflag3          // NEON scalar index plus 1 (Dd[x])


o_text = ida_ua.o_idpspec5    # Arbitrary text stored in the operand
                              # structure starting at the 'value' field
                              # up to 16 bytes (with terminating zero)
o_cond = ida_ua.o_idpspec5+1  # ARM condition as an operand
                              # condition is stored in 'value' field


# bits stored in specflag1 for APSR register
APSR_nzcv =       0x01
APSR_q =          0x02
APSR_g =          0x04
# for SPSR/CPSR
CPSR_c =          0x01
CPSR_x =          0x02
CPSR_s =          0x04
CPSR_f =          0x08
# for banked registers (R8-R12, SP, LR/ELR, SPSR), this flag is set
BANKED_MODE =     0x80  # the mode is in low 5 bits (arm_mode_t)


# Shift types:
class shift_t(IntEnum):
    LSL = 0               # logical left         LSL #0 - don't shift
    LSR = auto()          # logical right        LSR #0 means LSR #32
    ASR = auto()          # arithmetic right     ASR #0 means ASR #32
    ROR = auto()          # rotate right         ROR #0 means RRX
    RRX = auto()          # extended rotate right

    # ARMv8 shifts
    MSL = auto()          # masked shift left (ones are shifted in from the right)

    # extending register operations
    UXTB = auto()
    UXTH = auto()
    UXTW = auto()
    UXTX = auto()         # alias for LSL
    SXTB = auto()
    SXTH = auto()
    SXTW = auto()
    SXTX = auto()


class RegNo(IntEnum):
    R0 = 0
    R1 = auto()
    R2 = auto()
    R3 = auto()
    R4 = auto()
    R5 = auto()
    R6 = auto()
    R7 = auto()
    R8 = auto()
    R9 = auto()
    R10 = auto()
    R11 = auto()
    R12 = auto()
    R13 = auto()
    R14 = auto()
    R15 = auto()

    CPSR = auto()
    CPSR_flg = auto()

    SPSR = auto()
    SPSR_flg = auto()

    T = auto()
    rVcs = auto()
    rVds = auto()   # virtual registers for code and data segments
    Racc0 = auto()  # Intel xScale coprocessor accumulator

    # VFP system registers
    FPSID = auto()
    FPSCR = auto()
    FPEXC = auto()
    FPINST = auto()
    FPINST2 = auto()
    MVFR0 = auto()
    MVFR1 = auto()

    # msr system registers
    SYSM_APSR = auto()
    SYSM_IAPSR = auto()
    SYSM_EAPSR = auto()
    SYSM_XPSR = auto()
    SYSM_IPSR = auto()
    SYSM_EPSR = auto()
    SYSM_IEPSR = auto()
    SYSM_MSP = auto()
    SYSM_PSP = auto()
    SYSM_PRIMASK = auto()
    SYSM_BASEPRI = auto()
    SYSM_BASEPRI_MAX = auto()
    SYSM_FAULTMASK = auto()
    SYSM_CONTROL = auto()

    Q0 = auto()
    Q1 = auto()
    Q2 = auto()
    Q3 = auto()
    Q4 = auto()
    Q5 = auto()
    Q6 = auto()
    Q7 = auto()
    Q8 = auto()
    Q9 = auto()
    Q10 = auto()
    Q11 = auto()
    Q12 = auto()
    Q13 = auto()
    Q14 = auto()
    Q15 = auto()

    D0 = auto()
    D1 = auto()
    D2 = auto()
    D3 = auto()
    D4 = auto()
    D5 = auto()
    D6 = auto()
    D7 = auto()
    D8 = auto()
    D9 = auto()
    D10 = auto()
    D11 = auto()
    D12 = auto()
    D13 = auto()
    D14 = auto()
    D15 = auto()
    D16 = auto()
    D17 = auto()
    D18 = auto()
    D19 = auto()
    D20 = auto()
    D21 = auto()
    D22 = auto()
    D23 = auto()
    D24 = auto()
    D25 = auto()
    D26 = auto()
    D27 = auto()
    D28 = auto()
    D29 = auto()
    D30 = auto()
    D31 = auto()

    S0 = auto()
    S1 = auto()
    S2 = auto()
    S3 = auto()
    S4 = auto()
    S5 = auto()
    S6 = auto()
    S7 = auto()
    S8 = auto()
    S9 = auto()
    S10 = auto()
    S11 = auto()
    S12 = auto()
    S13 = auto()
    S14 = auto()
    S15 = auto()
    S16 = auto()
    S17 = auto()
    S18 = auto()
    S19 = auto()
    S20 = auto()
    S21 = auto()
    S22 = auto()
    S23 = auto()
    S24 = auto()
    S25 = auto()
    S26 = auto()
    S27 = auto()
    S28 = auto()
    S29 = auto()
    S30 = auto()
    S31 = auto()

    FIRST_FPREG = Q0
    LAST_FPREG = S31

    CF = auto()
    ZF = auto()
    NF = auto()
    VF = auto()

    # AArch64 registers
    # general-purpose registers
    X0 = auto()
    X1 = auto()
    X2 = auto()
    X3 = auto()
    X4 = auto()
    X5 = auto()
    X6 = auto()
    X7 = auto()
    X8 = auto()
    X9 = auto()
    X10 = auto()
    X11 = auto()
    X12 = auto()
    X13 = auto()
    X14 = auto()
    X15 = auto()
    X16 = auto()
    X17 = auto()
    X18 = auto()
    X19 = auto()
    X20 = auto()
    X21 = auto()
    X22 = auto()
    X23 = auto()
    X24 = auto()
    X25 = auto()
    X26 = auto()
    X27 = auto()
    X28 = auto()
    X29 = auto()
    XFP = X29  # frame pointer
    X30 = auto()
    XLR = X30  # link register
    XZR = auto()  # zero register (special case of GPR=31)
    XSP = auto()  # stack pointer (special case of GPR=31)
    XPC = auto()  # PC (not available as actual register)

    # 128-bit SIMD registers
    V0 = auto()
    V1 = auto()
    V2 = auto()
    V3 = auto()
    V4 = auto()
    V5 = auto()
    V6 = auto()
    V7 = auto()
    V8 = auto()
    V9 = auto()
    V10 = auto()
    V11 = auto()
    V12 = auto()
    V13 = auto()
    V14 = auto()
    V15 = auto()
    V16 = auto()
    V17 = auto()
    V18 = auto()
    V19 = auto()
    V20 = auto()
    V21 = auto()
    V22 = auto()
    V23 = auto()
    V24 = auto()
    V25 = auto()
    V26 = auto()
    V27 = auto()
    V28 = auto()
    V29 = auto()
    V30 = auto()
    V31 = auto()

    ARM_MAXREG = auto()  # must be the last entry


# ------------------------------------------------------------------
#      r0         *    argument word/integer result
#      r1-r3           argument word
#
#      r4-r8        S  register variable
#      r9           S  (rfp) register variable (real frame pointer)
#
#      r10        F S  (sl) stack limit (used by -mapcs-stack-check)
#      r11        F S  (fp) argument pointer
#      r12             (ip) temp workspace
#      r13        F S  (sp) lower end of current stack frame
#      r14             (lr) link address/workspace
#      r15        F    (pc) program counter
#
#      f0              floating point result
#      f1-f3           floating point scratch
#
#      f4-f7        S  floating point variables
PC =      RegNo.R15
LR =      RegNo.R14
SP =      RegNo.R13
FP =      RegNo.R11
FP2 =     RegNo.R7  # in thumb mode


def getreg(op: ida_ua.op_t) -> int:
    if (
        op.type == ida_ua.o_reg
        or op.type == o_shreg
        and op.specflag2 == shift_t.LSL  # specflag2 = shift type
        and op.value == 0   # value = shift count
    ):
        return op.reg
    else:
        return -1


def isreg(op: ida_ua.op_t, reg: int) -> bool:
    return getreg(op) == reg


def is_simple_phrase(insn: ida_ua.insn_t, op: ida_ua.op_t) -> bool:
    """
    is it simply [Rx, Ry]?
    no shift, no negation, no post-index, no writeback
    """
    return (
        op.type == ida_ua.o_phrase
        and op.specflag2 == shift_t.LSL
        and op.value == 0
        and (insn.auxpref & (aux_negoff | aux_postidx | aux_wback)) == 0
    )


# ------------------------------------------------------------------
# Condition codes:
class cond_t(IntEnum):
    cEQ = 0               # 0000 Z                        Equal
    cNE = auto()          # 0001 !Z                       Not equal
    cCS = auto()          # 0010 C                        Unsigned higher or same
    cCC = auto()          # 0011 !C                       Unsigned lower
    cMI = auto()          # 0100 N                        Negative
    cPL = auto()          # 0101 !N                       Positive or Zero
    cVS = auto()          # 0110 V                        Overflow
    cVC = auto()          # 0111 !V                       No overflow
    cHI = auto()          # 1000 C & !Z                   Unsigned higher
    cLS = auto()          # 1001 !C | Z                   Unsigned lower or same
    cGE = auto()          # 1010 (N & V) | (!N & !V)      Greater or equal
    cLT = auto()          # 1011 (N & !V) | (!N & V)      Less than
    cGT = auto()          # 1100 !Z & ((N & V)|(!N & !V)) Greater than
    cLE = auto()          # 1101 Z | (N & !V) | (!N & V)  Less than or equal
    cAL = auto()          # 1110 Always
    cNV = auto()          # 1111 Never
    cLAST = auto()


def get_cond(insn: ida_ua.insn_t) -> cond_t:
    # condition code of instruction will be kept in segpref
    return cond_t(insn.segpref)


def has_cond(insn: ida_ua.insn_t) -> bool:
    return insn.segpref != cond_t.cAL


def is_negated_cond(cond: cond_t) -> bool:
    return (cond & cond_t.cNE) != 0


def invert_cond(cond: cond_t) -> cond_t:
    if cond < cond_t.cLAST:
        return cond_t(cond ^ 1)
    return cond_t.cLAST


# ----------------------------------------------------------------------
# see ARMExpandImm_C/ThumbExpandImm_C in ARM ARM
def may_set_carry(itype: int) -> bool:
    return itype in (
        ida_allins.ARM_and,
        ida_allins.ARM_bic,
        ida_allins.ARM_eor,
        ida_allins.ARM_mov,
        ida_allins.ARM_mvn,
        ida_allins.ARM_orn,
        ida_allins.ARM_orr,
        ida_allins.ARM_teq,
        ida_allins.ARM_tst,
    )


# ----------------------------------------------------------------------
# if true, then ASPR.C is set to bit 31 of the immediate constant
def imm_sets_carry(insn: ida_ua.insn_t) -> bool:
    itype = insn.itype

    if itype in (
        ida_allins.ARM_and,
        ida_allins.ARM_bic,
        ida_allins.ARM_eor,
        ida_allins.ARM_mov,
        ida_allins.ARM_mvn,
        ida_allins.ARM_orn,
        ida_allins.ARM_orr,
    ):
        # flags are updated if S suffix is used
        return (insn.auxpref & (aux_immcarry|aux_cond)) == (aux_immcarry|aux_cond)
    elif itype in (
        ida_allins.ARM_teq,
        ida_allins.ARM_tst,
    ):
        # these two always update flags
        return (insn.auxpref & aux_immcarry) != 0
    else:
        return False

